// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"bytes"
	"context"
	"crypto/tls"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"os"
	"os/signal"
	"path"
	"strings"
	"syscall"
	"time"

	"github.com/gorilla/websocket"
	gw "github.com/orus-io/nats-websocket-gw"
	"github.com/spf13/cobra"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"

	xbushttp "xbus.io/xbus/http"
)

const (
	flagHTTPHost = "http.host"
	flagHTTPPort = "http.port"

	flagHTTPWebSocket = "http.websocket"
	flagHTTPUIPath    = "http.uipath"
)

// serveCmd represents the "serve" command.
var serveCmd = &cobra.Command{
	Use:   "serve",
	Short: "Serve http server.",
	Run: func(_ *cobra.Command, _ []string) {
		os.Exit(serve())
	},
}

func serve() int {
	configDir := path.Dir(v.ConfigFileUsed())
	if _, err := xbus.LoadTLSOptionFiles(v, v.ConfigFileUsed(), Logger); err != nil {
		Logger.Error().Msgf("%s", err)
		return 1
	}

	opts := xbus.COptionsFromViper(v, api.Account_GATEWAY, configDir, Logger)

	g := xbushttp.NewGateway(opts, Logger, LoggerOutput)

	uiPath := v.GetString(flagHTTPUIPath)
	uiIndexPath := path.Join(uiPath, "index.html")
	uiEnabled := uiPath != ""
	if uiEnabled {
		g.Router.PathPrefix("/ui").HandlerFunc(
			func(w http.ResponseWriter, r *http.Request) {
				upath := strings.TrimPrefix(r.URL.Path, "/ui")
				if upath == "" {
					newPath := "/ui/"
					if q := r.URL.RawQuery; q != "" {
						newPath += "?" + q
					}
					w.Header().Set("Location", newPath)
					w.WriteHeader(http.StatusMovedPermanently)
					return
				}
				if !strings.HasPrefix(upath, "/") {
					upath = "/" + upath
					r.URL.Path = upath
				}
				fpath := path.Join(uiPath, upath)
				if _, err := os.Stat(fpath); os.IsNotExist(err) {
					http.ServeFile(w, r, uiIndexPath)
				} else {
					http.ServeFile(w, r, fpath)
				}
			})

		g.Router.Path("/").HandlerFunc(
			func(w http.ResponseWriter, _ *http.Request) {
				w.Header().Set("Location", "/ui/")
				w.WriteHeader(http.StatusTemporaryRedirect)
			})
	}
	if v.GetBool(flagHTTPWebSocket) || uiEnabled {
		Logger.Info().Msgf("Enabling websocket gateway")

		settings := gw.Settings{
			NatsAddr:  strings.TrimPrefix(opts.ConnOptions.Url, "nats://"),
			EnableTLS: true,
			TLSConfig: &tls.Config{
				InsecureSkipVerify: true, // TODO load the actual RootCAs
				Certificates: []tls.Certificate{{
					Certificate: [][]byte{opts.TLS.Certificate},
					PrivateKey:  opts.TLS.PrivKey,
				}},
			},
			Trace: v.GetBool("logging.trace"),

			ConnectHandler: func(natsConn *gw.NatsConn, r *http.Request, ws *websocket.Conn) error {
				socketMode := websocket.TextMessage
				if mode := r.URL.Query()["mode"]; len(mode) == 1 && mode[0] == "binary" {
					socketMode = websocket.BinaryMessage
				}
				// send the server info
				infoCmd := bytes.Join(
					[][]byte{
						[]byte("INFO "),
						[]byte(natsConn.ServerInfo),
						[]byte("\r\n"),
					}, nil)
				if err := ws.WriteMessage(socketMode, infoCmd); err != nil {
					return err
				}

				// wait for the CONNECT from the client
				_, connectCmd, err := ws.ReadMessage()
				if err != nil {
					return err
				}
				splitted := bytes.SplitN(connectCmd, []byte(" "), 2)
				if len(splitted) != 2 {
					return fmt.Errorf("Invalid command: '%s'", connectCmd)
				}
				if !bytes.Equal(splitted[0], []byte("CONNECT")) {
					if bytes.Equal(splitted[0], []byte("-ERR")) {
						if err := ws.WriteMessage(socketMode, connectCmd); err != nil {
							Logger.Error().Msgf("error writing to the websocket: %s", err)
						}
						return fmt.Errorf(
							"Server returned an error: %s",
							connectCmd)
					}
					return fmt.Errorf(
						"Invalid command. Expected 'CONNECT', got: %s",
						connectCmd)
				}

				var connectDict map[string]interface{}
				if err := json.Unmarshal(
					splitted[1][:len(splitted[1])-2],
					&connectDict,
				); err != nil {
					return fmt.Errorf("could not unmarshal: %w", err)
				}

				// Check that CONNECT options are not trying anything suspect
				// We must to that because the TLS cert of the GW is used to
				// connect to NATS, and that means a token could be forge to
				// authenticate without any extra credentials
				authToken := connectDict["auth_token"]
				if s, ok := authToken.(string); ok && strings.Contains(s, ":") {
					return fmt.Errorf("Invalid auth token. ':' is forbidden, got: %s", s)
				}
				// drop unauthorized auth values
				delete(connectDict, "jwt")
				delete(connectDict, "sig")

				cmd, err := json.Marshal(connectDict)
				if err != nil {
					return err
				}
				// pass it on
				cmd = bytes.Join([][]byte{
					[]byte("CONNECT "), cmd, []byte("\r\n"),
				}, nil)
				_, err = natsConn.Conn.Write(cmd)
				return err
			},
		}

		wsgw := gw.NewGateway(settings)
		g.Router.HandleFunc("/__raw__", wsgw.Handler)
	}

	if err := g.Startup(); err != nil {
		Logger.Error().Msgf("%s", err)
		return 1
	}
	defer g.Shutdown()

	addr := v.GetString(flagHTTPHost) + ":" + v.GetString(flagHTTPPort)
	Logger.Info().Msgf("Listening on %s", addr)

	if uiEnabled {
		urlhost := v.GetString(flagHTTPHost)
		if urlhost == "0.0.0.0" {
			urlhost = "127.0.0.1"
		}
		Logger.Info().Msgf(
			"Open http://%s:%d/ui/ in your brower",
			urlhost, v.GetInt(flagHTTPPort))
	}

	server := http.Server{Addr: addr, Handler: g.Router}

	signals := make(chan os.Signal, 1)
	signal.Notify(signals, os.Interrupt, syscall.SIGTERM)

	go func() {
		select {
		case <-signals:
			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()
			if err := server.Shutdown(ctx); !errors.Is(err, http.ErrServerClosed) {
				Logger.Err(err).Msg("error shutting down the http server")
			}
		case <-g.ClientDone():
			if g.XbusOptions.NoReconnect {
				ctx, cancel := context.WithTimeout(context.Background(), time.Second)
				defer cancel()
				if err := server.Shutdown(ctx); !errors.Is(err, http.ErrServerClosed) {
					Logger.Err(err).Msg("error shutting down the http server")
				}
			}
		}
	}()

	if err := server.ListenAndServe(); !errors.Is(err, http.ErrServerClosed) {
		Logger.Err(err).Msg("http server error")
	}
	return 0
}

func must(err error) {
	if err != nil {
		panic(err)
	}
}

func init() {
	flags := serveCmd.Flags()
	flags.String(flagHTTPHost, "0.0.0.0", "Listen on the given host/ip")
	flags.Int(flagHTTPPort, 8911, "Listen on the given port")

	flags.Bool(flagHTTPWebSocket, false, "Enable the websocket gateway")

	flags.String(flagHTTPUIPath, "", "Path to the ui static files.\nIf set, it is served on '/ui' and enables the websocket gateway")

	must(v.BindPFlag(flagHTTPHost, flags.Lookup(flagHTTPHost)))
	must(v.BindPFlag(flagHTTPPort, flags.Lookup(flagHTTPPort)))

	must(v.BindPFlag(flagHTTPWebSocket, flags.Lookup(flagHTTPWebSocket)))

	must(v.BindPFlag(flagHTTPUIPath, flags.Lookup(flagHTTPUIPath)))

	RootCmd.AddCommand(serveCmd)
}
