// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"errors"
	"os"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/utils"
)

// Arguments of the "accept" command.
const (
	acceptCmdArgFingerprint = iota
	acceptCmdArgCount
)

// accountAcceptCmd represents the account accept sub-command.
var accountAcceptCmd = &cobra.Command{
	Use:   "accept <identifier>",
	Short: "Accept an account.",
	Long:  "Accept a pending account request so it can be allowed to connect to Xbus.",
	Run:   acceptAccount,
}

// acceptAccount parses arguments to accept a pending account request.
func acceptAccount(cmd *cobra.Command, args []string) {
	if len(args) < acceptCmdArgCount {
		Logger.Error().Msgf("Not enough arguments; %d arguments are required.", acceptCmdArgCount)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	expire, err := getCertExpire(cmd)
	if err != nil {
		Logger.Error().Msgf("%s", err)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	os.Exit(doAcceptAccount(args[acceptCmdArgFingerprint], expire))
}

// doAcceptAccount accepts a pending account request.
func doAcceptAccount(identifier string, expire time.Time) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	id, err := findAccountID(conn, identifier)
	if errors.Is(err, xbus.ErrNotFound) {
		Logger.Error().Msgf("No such account: %s", identifier)
		return 1
	} else if err != nil {
		Logger.Error().Msgf("%s", err)
		return 1
	}

	// Got a match! Send an "accept" request.
	repAccount, err := conn.AccountAccept(id, expire)
	if err != nil {
		Logger.Error().Msgf("Could not accept the account %q: %s", id, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Error().Msgf("Could not accept the account %q: Still pending.", id)
		return 1
	}
	Logger.Info().Msgf("Successfully accepted the account %q.", id)
	return 0
}

func getCertExpire(cmd *cobra.Command) (time.Time, error) {
	expireString, err := cmd.Flags().GetString("cert-expire")
	if err != nil {
		return time.Time{}, err
	}
	return utils.ParseDateOrDuration(expireString)
}

func init() {
	accountCmd.AddCommand(accountAcceptCmd)

	flags := accountAcceptCmd.Flags()
	flags.String("cert-expire", "1y", "Expire date or duration for the account certificate")

	must(viper.BindPFlag("accountaccept.cert-expire", flags.Lookup("cert-expire")))
}
