// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"bufio"
	"fmt"
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
)

// Arguments of the "delete" command.
const (
	deleteCmdArgIdentifier = iota
	deleteCmdArgCount
)

// accountDeleteCmd represents the account delete sub-command.
var accountDeleteCmd = &cobra.Command{
	Use:   "delete <identifier>",
	Short: "Delete an account.",
	Long:  "Delete an account.",
	Run:   deleteAccount,
}

// deleteAccount parses arguments to delete an account
func deleteAccount(cmd *cobra.Command, args []string) {
	if len(args) < deleteCmdArgCount {
		Logger.Error().Msgf("Not enough arguments; %d arguments are required.", deleteCmdArgCount)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	os.Exit(doDeleteAccount(
		args[deleteCmdArgIdentifier],
		cmd.Flag("delete-actors").Value.String() == "true"))
}

// doDeleteAccount deletes an account
func doDeleteAccount(identifier string, deleteActors bool) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Error().Msgf("Error getting account %q: %s", identifier, err)
		return 1
	}

	actors, err := conn.ActorList([]*api.Actor{{AccountID: account.ID}})
	if err != nil {
		Logger.Error().Msgf("Error listing account %q actors: %s", identifier, err)
		return 1
	}

	if len(actors) != 0 && !deleteActors {
		// Get confirmation from the user

		fmt.Printf("The following actors will be deleted along with the account:\n")
		for _, actor := range actors {
			fmt.Printf(" - %s\n", actor.Name)
		}
		fmt.Printf("Do you confirm the deletion (y/N)? ")
		reader := bufio.NewReader(os.Stdin)
		text, _ := reader.ReadString('\n')
		if !(text == "Y\n" || text == "y\n" || text == "yes\n") {
			return 1
		}
	}

	// Got a match! Send a "delete" request.
	err = conn.AccountDelete(account.GetIDAsUUID())
	if err != nil {
		Logger.Error().Msgf("Could not delete the account %q: %s", account.Name, err)
		return 1
	}
	return 0
}

func init() {
	flags := accountDeleteCmd.Flags()
	flags.Bool("delete-actors", false, "Force delete the related actors")

	accountCmd.AddCommand(accountDeleteCmd)
}
