// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"fmt"
	"os"
	"strings"

	"github.com/spf13/cobra"
	yaml "gopkg.in/yaml.v2"
	"xbus.io/go-xbus/v4/api"
)

// accountExportCmd is the 'account export' sub-command
var accountExportCmd = &cobra.Command{
	Use:   "export <identifier>",
	Short: `Export an account configuration file.`,
	Long:  `To be really usable, the certificate will need to be renewed.`,
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			if err := cmd.Usage(); err != nil {
				Logger.Error().Msgf("Failed to print usage: %s", err)
			}
			os.Exit(1)
		}
		os.Exit(accountExport(args[0]))
	},
}

func accountExport(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Error().Msgf("Could not find account %s: %s", identifier, err)
		return 1
	}

	actors, err := conn.ActorList([]*api.Actor{{AccountID: account.ID}})
	if err != nil {
		Logger.Error().Msgf("Could not load actors of account %s: %s", identifier, err)
		return 1
	}

	var actorsOutput yaml.MapSlice
	for i, actor := range actors {
		key := actor.Name
		conf := yaml.MapSlice{
			{Key: "kind", Value: strings.ToLower(actor.Kind.String())},
			{Key: "roles", Value: actor.Roles},
		}
		if strings.Contains(key, ".") {
			key = fmt.Sprintf("%d", i)
			conf = append(conf, yaml.MapItem{Key: "name", Value: actor.Name})
		}
		actorsOutput = append(actorsOutput, yaml.MapItem{
			Key:   key,
			Value: conf,
		})
	}

	output := yaml.MapSlice{
		{Key: "account-name", Value: account.Name},
		{Key: "nats-url", Value: conn.GetConn().ConnectedUrl()},
		{Key: "actors", Value: actorsOutput},
	}
	b, err := yaml.Marshal(output)
	if err != nil {
		Logger.Error().Msgf("Error marshaling the account data: %s", err)
		return 1
	}
	os.Stdout.Write(b)
	return 0
}

func init() {
	accountCmd.AddCommand(accountExportCmd)
}
