// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"errors"
	"os"

	"github.com/spf13/cobra"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// Arguments of the "reject" command.
const (
	rejectCmdArgFingerprint = iota
	rejectCmdArgCount
)

// accountRejectCmd represents the account reject sub-command.
var accountRejectCmd = &cobra.Command{
	Use:   "reject <identifier>",
	Short: "Reject an account.",
	Long:  "Reject a pending account request; it will not be allowed to connect to Xbus.",
	Run:   rejectAccount,
}

// rejectAccount parses arguments to reject a pending account request.
func rejectAccount(cmd *cobra.Command, args []string) {
	if len(args) < rejectCmdArgCount {
		Logger.Error().Msgf("Not enough arguments; %d arguments are required.", rejectCmdArgCount)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	os.Exit(doRejectAccount(args[rejectCmdArgFingerprint]))
}

// doRejectAccount rejects a pending account request.
func doRejectAccount(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	id, err := findAccountID(conn, identifier)
	if errors.Is(err, xbus.ErrNotFound) {
		Logger.Error().Msgf("No such account: %s", identifier)
		return 1
	} else if err != nil {
		Logger.Error().Msgf("%s", err)
		return 1
	}

	// Got a match! Send a "reject" request.
	repAccount, err := conn.AccountReject(id)
	if err != nil {
		Logger.Error().Msgf("Could not reject the account %q: %s", id, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Error().Msgf("Could not reject the account %q: Still pending.", id)
		return 1
	}
	Logger.Info().Msgf("Successfully rejected the account %q.", id)
	return 0
}

func init() {
	accountCmd.AddCommand(accountRejectCmd)
}
