// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api/control"
)

// accountResetCertCmd is 'account reset-cert' sub-command.
var accountResetCertCmd = &cobra.Command{
	Use:   "reset-cert <identifier>",
	Short: "Reset an account certificate (/!\\ temporarily disable the account /!\\)",
	Long: `Disable an account and set its certificate as renewable,
so the client may submit a new CSR, which then has to be accepted.`,
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			if err := cmd.Usage(); err != nil {
				Logger.Error().Msgf("Failed to print usage: %s", err)
			}
			os.Exit(1)
		}
		os.Exit(accountResetCert(args[0]))
	},
}

func accountResetCert(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Error().Msgf("Could not find account %s: %s", identifier, err)
		return 1
	}

	// TODO add a confirmation message
	if err := conn.Control.Account.ResetCert(
		&control.AccountIDRequest{AccountID: account.ID},
	); err != nil {
		Logger.Error().Msgf("Failed: %s", err)
	}
	Logger.Info().Msgf("The account %s is now ready for a certificate renewal.", account.Name)
	return 0
}

func init() {
	accountCmd.AddCommand(accountResetCertCmd)
}
