// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"fmt"
	"os"
	"sort"
	"strings"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
)

// actorListCmd represents the list command
var actorListCmd = &cobra.Command{
	Use:   "list",
	Short: "",
	Long:  ``,
	Run: func(cmd *cobra.Command, _ []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		id, _ := api.UUIDFromString(cmd.Flag("id").Value.String())
		actor := api.Actor{
			ID:    id.ToBytes(),
			Name:  cmd.Flag("name").Value.String(),
			Roles: []string{cmd.Flag("role").Value.String()},
		}
		if cmd.Flag("type").Value.String() != "" {
			err = actor.Kind.UnmarshalText([]byte(cmd.Flag("type").Value.String()))
			if err != nil {
				Logger.Error().Msgf("Error parsing type. Got %s", err)
				return
			}
		}
		actorList := []*api.Actor{}
		if !actor.GetIDAsUUID().IsZero() || actor.Name != "" || !actor.GetAccountIDAsUUID().IsZero() {
			actorList = append(actorList, &actor)
		}
		actors, err := c.ActorList(actorList)
		if err != nil {
			Logger.Error().Msgf("Error: %s", err)
			return
		}
		if len(actors) == 0 {
			fmt.Println("No actors found")
		} else {
			accountList, err := c.AccountList()
			if err != nil {
				Logger.Error().Msgf("Could not load the accounts: %s", err)
			}
			var (
				accountsByUUID = make(map[api.UUID]*api.Account)
			)
			for _, account := range accountList {
				accountsByUUID[account.GetIDAsUUID()] = account
			}
			table := tablewriter.NewWriter(os.Stdout)
			headers := []string{"id", "account", "name", "kind", "roles", "status", "state"}
			table.SetHeader(headers)
			table.SetBorder(false)
			table.SetHeaderLine(false)
			table.SetColumnSeparator(" ")

			sort.Slice(actors, func(i, j int) bool {
				accountI := accountsByUUID[actors[i].GetAccountIDAsUUID()].Name
				if accountI == "" {
					accountI = actor.GetAccountIDAsUUID().String()
				}
				accountJ := accountsByUUID[actors[j].GetAccountIDAsUUID()].Name
				if accountJ == "" {
					accountJ = actor.GetAccountIDAsUUID().String()
				}
				if c := strings.Compare(accountI, accountJ); c != 0 {
					return c < 0
				}
				return strings.Compare(actors[i].Name, actors[j].Name) < 0
			})

			for _, actor := range actors {
				state := ""
				if actor.Kind != api.Actor_EMITTER {
					if actor.Online {
						state = "online"
						if actor.Unresponsive {
							state += fmt.Sprintf(", unresponsive (last seen: %v)", actor.LastSeen)
						}
					} else {
						state = "offline"
					}
				}
				table.Append([]string{
					actor.GetIDAsUUID().String(),
					accountsByUUID[actor.GetAccountIDAsUUID()].Name,
					actor.Name,
					actor.Kind.String(),
					strings.Join(actor.Roles, "\n"),
					actor.GetStatus().String(),
					state,
				})
			}

			table.Render()
		}
	},
}

func init() {
	actorCmd.AddCommand(actorListCmd)
}
