// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"bytes"
	"io"
)

// Indent a block of text
func Indent(data, indent []byte) []byte {
	b := bytes.Buffer{}
	w := IndentWriter{indent, &b, false}
	_, _ = w.Write(data)
	return b.Bytes()
}

// IndentString a block of text
func IndentString(data, indent string) string {
	b := bytes.Buffer{}
	w := IndentWriter{[]byte(indent), &b, false}
	_, _ = w.Write([]byte(data))
	return b.String()
}

// IndentWriter writes with added indentation
type IndentWriter struct {
	indent      []byte
	out         io.Writer
	linestarted bool
}

func (w *IndentWriter) Write(data []byte) (int, error) {
	lines := bytes.Split(data, []byte("\n"))
	if !w.linestarted {
		if _, err := w.out.Write(w.indent); err != nil {
			return 0, err
		}
	}
	if n, err := w.out.Write(append(lines[0], '\n')); err != nil {
		return n, err
	}
	total := len(lines[0])
	for _, line := range lines[1 : len(lines)-1] {
		if _, err := w.out.Write(w.indent); err != nil {
			return total, err
		}
		n, err := w.out.Write(append(line, '\n'))
		total += n
		if err != nil {
			return total, err
		}
	}
	lastline := lines[len(lines)-1]
	if len(lastline) == 0 {
		w.linestarted = false
	} else {
		w.linestarted = true
		if _, err := w.out.Write(w.indent); err != nil {
			return total, err
		}
		n, err := w.out.Write(lastline)
		total += n
		if err != nil {
			return total, err
		}
	}
	return total, nil
}
