// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"io"
	"os"

	"github.com/spf13/cobra"
	yaml "gopkg.in/yaml.v2"
	xbus "xbus.io/go-xbus/v4"

	"xbus.io/xbus/storage"
)

var pipelineGetCmd = &cobra.Command{
	Use:   "get <name or id> [version]",
	Short: "Get a pipeline definition",
	Long:  "Get a pipeline definition (its graph) and write it to stdout",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) == 0 || len(args) > 2 {
			usage(cmd)
		}
		c, err := GetClient(false)
		if err != nil {
			Logger.Error().Msgf("%s", err)
			os.Exit(-1)
		}

		var (
			identifier = args[0]
			version    string
			filename   = cmd.Flag("output").Value.String()
		)
		if len(args) == 2 {
			version = args[1]
		}
		var out io.Writer
		if filename == "-" {
			out = os.Stdout
		} else {
			f, err := os.Create(filename)
			if err != nil {
				Logger.Error().Msgf("Error opening output file: %s", err)
				os.Exit(-1)
			}
			defer func() {
				if err := f.Close(); err != nil {
					Logger.Error().Msgf("Error saving pipeline: %s", err)
					os.Exit(-1)
				}
			}()
			out = f
		}
		if err := pipelineGet(c, identifier, version, out); err != nil {
			Logger.Error().Msgf("Error getting pipeline: %s", err)
			os.Exit(-1)
		}
	},
}

func pipelineGet(c *xbus.Client, identifier, version string, out io.Writer) error {
	info := pipelineInfo(identifier, version)

	str, err := c.PipelineGet(info)
	if err != nil {
		return err
	}

	var graph storage.Graph
	if err := yaml.Unmarshal([]byte(str), &graph); err != nil {
		return err
	}
	sortNodeSlice(graph.Nodes)
	sortEdgeSlice(graph.Edges)
	b, err := yaml.Marshal(graph)
	if err != nil {
		return err
	}

	if _, err := out.Write(b); err != nil {
		return err
	}
	return nil
}

func init() {
	pipelineCmd.AddCommand(pipelineGetCmd)
	pipelineGetCmd.Flags().StringP("output", "o", "-", "Target filename (default to stdout)")
}
