// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"bytes"
	"os"
	"strings"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
	yaml "gopkg.in/yaml.v2"

	"xbus.io/xbus/storage"
)

var pipelineListCmd = &cobra.Command{
	Use:   "list [name]",
	Short: "List pipelines, optionnaly filtered by name",
	Run:   pipelineList,
}

func pipelineList(cmd *cobra.Command, args []string) {
	c, err := GetClient(false)
	if err != nil {
		return
	}
	if len(args) > 1 {
		usage(cmd)
	}

	extended := cmd.Flag("extended").Value.String() == "true"

	filter, err := parsePipelineFilterFlags(cmd)
	if err != nil {
		Logger.Error().Msgf("%s", err)
		return
	}

	pipelines, graphs, err := filterPipelineList(c, filter)
	if err != nil {
		Logger.Error().Msgf("%s", err)
		return
	}

	table := tablewriter.NewWriter(os.Stdout)
	headers := []string{"id", "pipeline", "version", "status"}
	if extended {
		headers = append(
			headers, "type", "sourcematch", "emitter(s)", "workers(s)", "consumer(s)")
	}
	table.SetHeader(headers)
	if !extended {
		table.SetBorder(false)
		table.SetHeaderLine(false)
		table.SetColumnSeparator(" ")
	} else {
		table.SetRowLine(true)
	}
	table.SetAlignment(tablewriter.ALIGN_LEFT)

	for i, info := range pipelines {
		var graph *storage.Graph
		if extended {
			if graphs != nil {
				graph = graphs[i]
			}
			if graph == nil {
				b := bytes.NewBuffer(nil)
				if err := pipelineGet(c, info.Name, info.Version, b); err != nil {
					Logger.Error().Msgf("%s", err)
					return
				}
				graph = &storage.Graph{}
				if err := yaml.Unmarshal(b.Bytes(), graph); err != nil {
					Logger.Error().Msgf("%s", err)
					return
				}
			}
		}
		version := info.Version
		if version == "" {
			version = "<none>"
		}
		values := []string{
			info.GetIDAsUUID().String(),
			info.Name,
			version,
			info.Status.String(),
		}
		if extended {
			var (
				sourcematch, emitters, workers, consumers []string
				requestreply                              bool
			)
			for _, node := range graph.Nodes {
				if node.Type == storage.RequestorNode {
					requestreply = true
				}
				switch node.Type {
				case storage.EmitterNode, storage.RequestorNode:
					if node.SourceMatch != nil {
						sourcematch = append(sourcematch, node.SourceMatch.EventTypes...)
					}
					for _, id := range node.ActorIDs {
						emitters = append(emitters, id.String())
					}
					emitters = append(emitters, node.Actors...)
					for _, role := range node.Roles {
						emitters = append(emitters, "role:"+role)
					}
				case storage.WorkerNode:
					for _, id := range node.ActorIDs {
						workers = append(workers, id.String())
					}
					workers = append(workers, node.Actors...)
					for _, role := range node.Roles {
						workers = append(workers, "role:"+role)
					}
				case storage.ConsumerNode:
					for _, id := range node.ActorIDs {
						consumers = append(consumers, id.String())
					}
					consumers = append(consumers, node.Actors...)
					for _, role := range node.Roles {
						consumers = append(consumers, "role:"+role)
					}
				}
			}
			if requestreply {
				values = append(values, "request/reply")
			} else {
				values = append(values, "one way")
			}
			values = append(values,
				strings.Join(sourcematch, "\n"),
				strings.Join(emitters, "\n"),
				strings.Join(workers, "\n"),
				strings.Join(consumers, "\n"))
		}
		table.Append(values)
	}
	table.Render()
}

func init() {
	pipelineListCmd.Flags().Bool("extended", false, "Print more information on the pipelines")
	addPipelineFilterFlags(pipelineListCmd.Flags(), false)
	pipelineCmd.AddCommand(pipelineListCmd)
}
