// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"fmt"
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
)

var psPurgeCmd = &cobra.Command{
	Use:   "purge",
	Short: "Purge process data",
	Long: `Remove all the data related to some process(es).
Defaults to all the processes terminated more than 1 month ago`,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{
				IncludeTerminatedStatuses: true,
				StatusChangedTo:           "31d", // Approx 1 month
			})
		if err != nil {
			Logger.Error().Msgf("Invalid filter: %s", err)
			return
		}
		confirm := "Do you confirm the definitive deletion of %d process(es) (there is NO going back!) (yes/NO)?"
		if cmd.Flag("yes").Value.String() == "true" {
			confirm = ""
		}
		list, err := getProcessList(c, filter, confirm, false)
		if err != nil {
			fmt.Println(err)
			os.Exit(1)
		}
		if len(list) == 0 {
			fmt.Println("No process matched the filter")
			os.Exit(0)
		}
		var (
			pids     []api.UUID
			pidbatch [][]api.UUID
		)
		for _, p := range list {
			pids = append(pids, p.GetIDAsUUID())
			if len(pids) >= 10000 {
				pidbatch = append(pidbatch, pids)
				pids = nil
			}
		}
		if len(pids) > 0 {
			pidbatch = append(pidbatch, pids)
		}
		var (
			total         = len(list)
			offset uint32 = 0
		)
		ctx, _ := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)
		fmt.Print("\nPurge: ")
		for _, batch := range pidbatch {
			if err := c.ProcessPurge(ctx, batch, func(pr uint32, _ uint32) {
				fmt.Printf("\rPurge: %d/%d", offset+pr, total)
			}); err != nil {
				fmt.Print("\n")
				Logger.Error().Msgf("Error: %s", err)
				return
			}
			offset += uint32(len(batch))
		}
		fmt.Print("\n")
		Logger.Info().Msgf("Purged %d process(es)", offset)
	},
}

func init() {
	psCmd.AddCommand(psPurgeCmd)
}
