// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"fmt"
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
)

var psReplayCmd = &cobra.Command{
	Use:   "replay",
	Short: "Replay a process",
	Run: func(cmd *cobra.Command, args []string) {
		yes := cmd.Flag("yes").Value.String() == "true"
		match := cmd.Flag("match").Value.String() == "true"

		c, err := GetClient(false)
		if err != nil {
			fmt.Println(err)
			os.Exit(1)
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{})
		if err != nil {
			Logger.Error().Msgf("Invalid filter: %s", err)
			os.Exit(1)
		}
		msg := ""
		if !yes {
			msg = "Will replay %d process(es). Do you confirm (yes/NO)?"
		}
		list, err := getProcessList(c, filter, msg, false)
		if err != nil {
			fmt.Println(err)
			os.Exit(1)
		}
		if len(list) == 0 {
			fmt.Println("No process matched the filter")
			os.Exit(0)
		}
		errorCount := 0
		for i, p := range list {
			process, err := c.Control.Process.Replay(
				&control.ProcessReplayRequest{
					Process: &api.Process{ID: p.ID},
					Match:   match,
				},
			)
			if err != nil {
				errorCount++
				Logger.Error().Msgf("Process %s: %s", p.ID, err)
			} else {
				Logger.Info().
					Str("processid", p.GetIDAsUUID().String()).
					Str("new-processid", process.GetIDAsUUID().String()).
					Msg("process replay triggered")
			}
			fmt.Printf("\r%d success, %d errors / %d     ", (i + 1 - errorCount), errorCount, len(list))
		}
		fmt.Printf("\n%d success, %d errors", len(list)-errorCount, errorCount)
		if errorCount != 0 {
			Logger.Error().Msgf("Encountered %d errors, some process could not be replayed", errorCount)
			os.Exit(1)
		}
	},
}

func init() {
	psCmd.AddCommand(psReplayCmd)

	flags := psReplayCmd.PersistentFlags()
	flags.Bool("match", false, "Match again to find the right pipeline. If not, use the same pipeline as before, even if inactive or if there is a better match")
}
