// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"fmt"

	"github.com/orus-io/qb"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql"
	"xbus.io/xbus/storage/sql/model"
)

// accountCmd represents the account command.
var accountCmd = &cobra.Command{
	Use:   "account",
	Short: "Manage accounts that are allowed to connect to Xbus.",
	Long: `This command contains several sub-commands to manage accounts and the
permissions they have within Xbus.

In particular, it is possible from this interface to accept / reject pending
account requests.

Note that this is usually done via the "xbusctl" program, although the commands
are available here as well.`,
	PersistentPreRun: preRunCheckCertCertDir,
}

// findAccount finds a pending account with the specified fingerprint or name or id
func findAccount(
	ctx context.Context, accountStorage storage.AccountStorage, key string,
) (
	account *api.Account, found bool,
) {
	var ok bool
	id, err := api.UUIDFromString(key)
	switch {
	case err == nil:
		account, ok = accountStorage.Get(ctx, id)
		if !ok {
			Logger.Error().Msgf("No account by this ID: %s", id)
			return
		}
	case len(key) == 59:
		for _, a := range accountStorage.List(ctx) {
			// TODO To be done elsewhere? Should fingerprints be cached?
			if tools.GetCSRHash(a.CSR, true) == key {
				account = a
				ok = true
				break
			}
		}
		if !ok {
			Logger.Error().Msgf("No account by this fingerprint: %s", key)
			return
		}
	default:
		account, ok = accountStorage.GetByName(ctx, key)
		if !ok {
			Logger.Error().Msgf("No account by this name: %s", key)
			return
		}
	}

	return account, true
}

func fmtAccount(account *api.Account, csrHash bool, certDetails bool) string {
	formattedAccount := fmt.Sprintf(`
ID: %s
Name: %s
Type: %s
Status: %s`,
		account.ID,
		account.Name,
		account.Type,
		account.Status,
	)
	if csrHash {
		formattedAccount += fmt.Sprintf(`
CSR fingerprint: %s`,
			tools.GetCSRHash(account.CSR, true),
		)
	}
	if certDetails {
		formattedAccount += fmt.Sprintf(`
CSR: %s
Cert: %s`,
			account.CSR,
			account.Cert,
		)
	}
	return formattedAccount
}

// getAccountStorage provides access to the DB account storage.
func getAccountStorage() (accountStorage storage.AccountStorage) {
	m, db, err := model.NewDB("postgres",
		viper.GetString("database.dsn"),
		viper.GetInt("database.max-connections"),
		Logger, qb.LDefault)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the database: %s", err)
		return nil
	}
	return sql.NewAccountStorage(m, db, Logger)
}

func init() {
	RootCmd.AddCommand(accountCmd)
}
