// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"crypto/x509"
	"encoding/pem"
	"os"

	"github.com/spf13/cobra"
)

// certGenerateCmd represents the certificate generation sub-command.
var certGenerateCmd = &cobra.Command{
	Use:   "generate",
	Short: "Generate certificates.",
	Long: "Generate certificates to be used by this program, based on private keys; see " +
		"sub-commands for details on the certificates that can be generated here. Also check " +
		"the \"key\" command for private key generation.",
}

// loadPrivKey loads the private key stored in the specified file.
func loadPrivKey(privKeyPath string) (privKey interface{}, err error) {
	privKeyData, err := os.ReadFile(privKeyPath)
	if err != nil {
		return nil, err
	}
	privKeyBlock, _ := pem.Decode(privKeyData)
	return x509.ParseECPrivateKey(privKeyBlock.Bytes)
}

// genCert is a helper that creates a certificate and saves it into a file.
func genCert(
	certPath string, privKeyPath string,
	genFunc func(priv interface{}) (certData []byte, err error),
) (generated bool) {
	// Load the private key.
	privKey, err := loadPrivKey(privKeyPath)
	if err != nil {
		Logger.Error().Msgf("Cannot load the %q private key", privKeyPath)
		return false
	}

	// Generate the certificate.
	certData, err := genFunc(privKey)
	if err != nil {
		Logger.Error().Msgf("Cannot create the certificate: %s", err)
		return false
	}
	if err := os.WriteFile(certPath, certData, 0644); err != nil {
		Logger.Error().Msgf("Cannot save the certificate to %q: %s", certPath, err)
		return false
	}

	Logger.Info().Msgf("Successfully generated %q.", certPath)
	return true
}

func init() {
	certCmd.AddCommand(certGenerateCmd)
}
