// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"os"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"xbus.io/xbus/utils"
)

// certGenerateServerCACmd represents the server CA generation sub-command.
var certGenerateServerCACmd = &cobra.Command{
	Use:   "server-ca",
	Short: "Generate certificates for the server CA.",
	Long: "Generate certificates for the server CA (the one clients connecting to Xbus will " +
		"check). The file will be named \"ServerCA.crt\" and is based on a \"ServerCA.key\" " +
		"private key.",
	Run: generateServerCA,
}

// generateServerCA parses arguments to generate the server CA.
func generateServerCA(cmd *cobra.Command, _ []string) {
	certDir := viper.GetString("cert-dir")
	hostDefs := cmd.Flag(flagHostDefs).Value.String()
	if hostDefs == "" {
		hostDefs = viper.GetString("nats.host")
	}
	expire, err := getExpire(cmd, "expire")
	if err != nil {
		Logger.Error().Msgf("Invalid --expire: %s", err)
		os.Exit(1)
		return
	}
	os.Exit(doGenerateServerCA(certDir, hostDefs, expire))
}

// doGenerateServerCA generates the server CA.
func doGenerateServerCA(certDir string, hostDefs string, expire time.Time) int {
	generated := genCert(
		certDir+"/ServerCA.crt", certDir+"/ServerCA.key",
		func(priv interface{}) ([]byte, error) {
			return utils.CreateServerRootCA(priv, time.Now(), expire, hostDefs)
		},
	)
	if !generated {
		return 1
	}
	return 0
}
func init() {
	certGenerateCmd.AddCommand(certGenerateServerCACmd)

	flags := certGenerateServerCACmd.Flags()
	flags.String(flagHostDefs, "", flagHostDefsHelp)
	flags.String(flagExpire, "1y", flagExpireHelp)
}
