// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package fullenv

import (
	"context"
	"errors"
	"regexp"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"
)

func newBasRelayActor(
	actor *xbus.Actor, _ xbus.Settings,
) (xbus.ActorService, error) {
	return &badRelayActor{actor}, nil
}

type badRelayActor struct {
	actor *xbus.Actor
}

func (a *badRelayActor) Process(apc *xbus.ActorProcessingContext) error {
	if _, err := apc.CompleteEnvelope(api.DefaultInput, time.Second); err != nil {
		apc.Log.Error().Msgf("Reception error: %s", err)
		return err
	}
	return errors.New("badRelayActor error")
}

func TestInvalidProcessState(t *testing.T) {
	fe, _ := NewFullEnvForTest(t, "invalid-process-state", "invalid_process_state.yaml")

	if err := fe.Init(); err != nil {
		t.Fatal(err)
	}
	defer fe.Close()

	fe.RegisterActorService("badrelayactor", newBasRelayActor)

	assert.NoError(t, fe.Startup())

	// Emit one message
	_, err := fe.Emitter.Actor.Emit(
		context.TODO(),
		envelope.MustNewEnvelope(
			envelope.MustNewTextMessage("msg", "hi"),
		),
	)
	require.NoError(t, err)

	// Wait for the process to be over
	time.Sleep(100 * time.Millisecond)

	// EXTRACT PIDS)
	re := regexp.MustCompile(`\| ([0-9a-f-]+) \| [^\|]+ \| [^\|]+ \| [^\|]+ \| [A-Z]+ +\|`)

	// The relay will misbehave, and the process should pass in a 'error' state
	matchLines, err := fe.XbusCtl.RunReadAndFindSubmatch(re, "ps", "--status-error")
	must(err)
	assert.Equal(t, 1, len(matchLines), "Should get 1 matching line, got:\n%v", matchLines)

	// Restarting xbus should not be an issue
}
