// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package fullstack

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/testutils"
)

func TestPasswordAuth(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	t.Run("Set_password", func(t *testing.T) {
		defer ft.SwitchT(t)()
		err := ft.Client.Control.Account.PasswordSet(
			&control.AccountPasswordSetRequest{
				AccountID: emitter.Client.ID().ToBytes(),
				Password:  "mypass",
			},
		)
		if err != nil {
			t.Error(err)
		}
	})

	t.Run("Password_login", func(t *testing.T) {
		defer ft.SwitchT(t)()
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = ""
		options.ConnOptions.User = emitter.Client.Name()
		options.ConnOptions.Password = "mypass"

		client := xbus.NewClient(options, ft.Logger.Logger, ft.Logger.LoggerOutput)

		require.NoError(t, client.Connect())
		assert.Equal(t, emitter.Client.ID(), client.ID())
	})

	t.Run("Invalid_password_login", func(t *testing.T) {
		defer ft.SwitchT(t)()
		ft.Logger.ExpectsError(testutils.ExpectsJSONContains(
			`{"username":"bob","message":"invalid password"}`,
		))
		ft.Logger.ExpectsError(testutils.ExpectsContains(
			"authentication error",
		))
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = ""
		options.ConnOptions.User = emitter.Client.Name()
		options.ConnOptions.Password = "wrong"

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		err := client.Connect()
		assert.EqualError(t, err, "nats: Authorization Violation")
	})
}

func TestSessionAuth(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	var (
		token  string
		token2 string
	)

	t.Run("Get a session token", func(t *testing.T) {
		if err := ft.Client.Control.Account.PasswordSet(
			&control.AccountPasswordSetRequest{
				AccountID: emitter.Client.ID().ToBytes(),
				Password:  "mypass",
			},
		); err != nil {
			t.Error(err)
		}

		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = ""
		options.ConnOptions.User = emitter.Client.Name()
		options.ConnOptions.Password = "mypass"

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		require.NoError(t, client.Connect())

		sessionToken, err := client.API.ClientAPI.GetSessionToken(
			&api.GetSessionTokenRequest{
				ExpiresIn: 1,
			},
		)
		if err != nil {
			t.Fatal(err)
		}
		token = sessionToken.Token
	})

	t.Run("Authenticate", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		assert.Equal(t, emitter.Client.ID(), client.ID())
	})

	t.Run("Renew token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		if emitter.Client.ID() != client.ID() {
			t.Fatalf(
				"Not authenticated with the expected account. Expected: %s, got: %s",
				emitter.Client.ID(), client.ID(),
			)
		}

		sessionToken, err := client.API.ClientAPI.GetSessionToken(&api.GetSessionTokenRequest{
			InvalidateToken: token,
			ExpiresIn:       1,
		})

		if err != nil {
			t.Fatal(err)
		}

		token2 = sessionToken.Token
	})

	t.Run("Renewed Token", func(t *testing.T) {
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token2

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		if err := client.Connect(); err != nil {
			t.Fatal(err)
		}
		defer client.Close()
		assert.Equal(t, emitter.Client.ID(), client.ID())
	})

	t.Run("Invalidated Token", func(t *testing.T) {
		defer ft.SwitchT(t)()
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		ft.Logger.ExpectsError(testutils.ExpectsContains("unknown session token"))
		ft.Logger.ExpectsError(testutils.ExpectsContains("authentication error"))

		if err := client.Connect(); err == nil {
			client.Close()
			t.Fatal("Client should not be able to connect with an invalidated token")
		}
	})

	t.Run("Expired Token", func(t *testing.T) {
		defer ft.SwitchT(t)()
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + token2

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		time.Sleep(time.Second * 3 / 2)

		ft.Logger.ExpectsError(testutils.ExpectsContains("session expired"))
		ft.Logger.ExpectsError(testutils.ExpectsContains("authentication error"))

		if err := client.Connect(); err == nil {
			client.Close()
			t.Fatal("Client should not be able to connect with an expired token")
		}
	})

	t.Run("Random token", func(t *testing.T) {
		defer ft.SwitchT(t)()
		options := xbus.COptions{
			Name:        emitter.Client.Name(),
			ConnOptions: ft.Server.NATSConnectOptions(),
		}

		options.ConnOptions.Token = "session#" + tools.RandomString(32)

		ft.Logger.ExpectsError(testutils.ExpectsContains("unknown session token"))
		ft.Logger.ExpectsError(testutils.ExpectsContains("authentication error"))

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		err := client.Connect()
		if err == nil {
			client.Close()
			t.Fatal("Client should not be able to connect with a random token")
		}
	})
}
