// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package process

import (
	"context"
	"fmt"
	"time"

	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/storage"
)

const envelopeExportChunkSize = 250 * 1024

// EnvelopeActorPair contains a pair (EnvelopeID, ActorID)
type EnvelopeActorPair struct {
	EnvelopeID api.UUID
	ActorID    api.UUID
}

// NewProcess creates a process
func NewProcess(
	actorStorage storage.ActorStorage,
	emitterID,
	envelopeID api.UUID,
	pipeline *pipeline.Pipeline,
	source storage.NodeOutputRef,
	manager *Manager,
) *Process {
	id, err := api.UUIDV4()
	if err != nil {
		panic(fmt.Sprintf("Cannot generate a UUID: %s", err))
	}
	return &Process{
		ID:                id,
		actors:            actorStorage,
		SourceRef:         source,
		TriggerEmitterID:  emitterID,
		Pipeline:          pipeline,
		TriggerEnvelopeID: envelopeID,
		manager:           manager,
	}
}

// Process orchestrate a pipeline execution
type Process struct {
	ID        api.UUID
	CreatedAt time.Time
	GroupID   api.UUID
	ReplayOf  api.UUID

	actors         storage.ActorStorage
	processStorage storage.ProcessStorage
	graphManager   *pipeline.Manager

	manager *Manager

	TriggerEmitterID  api.UUID
	TriggerEnvelopeID api.UUID

	Pipeline  *pipeline.Pipeline
	SourceRef storage.NodeOutputRef
}

// GetStatus returns the current status of the process
func (p Process) GetStatus(ctx context.Context) api.Process_Status {
	state := p.processStorage.GetState(ctx, p.ID)
	return state.Status
}

// Pause switch the process state to ProcessPaused
func (p Process) Pause(ctx context.Context) error {
	return p.processStorage.SetStatus(ctx, p.ID, api.Process_PAUSED, "")
}

// Resume switch the process state to ProcessRunning if previously paused
func (p Process) Resume(ctx context.Context) error {
	return p.processStorage.SetStatus(ctx, p.ID, api.Process_RUNNING, "")
}

// Cancel sets the process state to ERROR with "canceled: <reason>" reason
func (p Process) Cancel(ctx context.Context, reason string) error {
	return p.processStorage.SetStatus(ctx, p.ID, api.Process_ERROR, "Canceled: "+reason)
}

// MatchPipeline attempts to find the pipeline that will drive the process
func (p *Process) MatchPipeline(ctx context.Context, output string, eventTypes []string) error {
	var success bool
	// TODO UpdateReceiveStatus with ERROR if no match
	defer func() {
		if !success {
			p.manager.Log.Info().Msg("MatchPipeline failed, setting status to error")
			if err := p.processStorage.SetStatus(
				ctx, p.ID, api.Process_ERROR, "No matching pipeline",
			); err != nil {
				p.manager.Log.Err(err).Msg("Error setting process status")
			}
		}
	}()
	actor, ok := p.actors.Get(ctx, p.TriggerEmitterID)
	if !ok {
		return fmt.Errorf("Invalid actor")
	}
	pipe, outputRef, ok := p.graphManager.LookupPipeline(
		ctx, actor, output, eventTypes)
	if !ok {
		return fmt.Errorf("No matching pipeline")
	}
	p.Pipeline = pipe
	p.SourceRef = outputRef
	p.processStorage.SetPipeline(ctx, p.ID, p.Pipeline.GetIDAsUUID(), p.SourceRef)
	success = true
	return nil
}

// GetTargets returns the targets for an envelope coming from a given emitter
func (p Process) GetTargets(ctx context.Context, output storage.NodeOutputRef, envelopeID api.UUID, noRouteTableUpdate bool) ([]*api.EnvelopeTarget, error) {
	if p.Pipeline == nil {
		return nil, fmt.Errorf("Pipeline-less process")
	}
	// is the envelope already routed ?
	return p.processStorage.GetTargets(ctx, p.ID, output, envelopeID, noRouteTableUpdate,
		func() ([]*api.EnvelopeTarget, bool, error) {
			dests, err := p.Pipeline.GetDestinations(output)
			if err != nil {
				return nil, false, err
			}
			if len(dests) == 0 {
				panic("found no destination")
			}
			if len(dests) != 1 {
				panic("found more than one destination")
			}
			dest := dests[0]
			destNode, err := p.Pipeline.GetNode(dest.NodeID)
			if err != nil {
				return nil, false, fmt.Errorf("cannot find the target node: %w", err)
			}

			var actors []*api.Actor

			switch destNode.Type {
			case storage.ConsumerNode:
				if len(destNode.Roles) != 0 && len(destNode.Actors) != 0 {
					return nil, false, fmt.Errorf(
						"consumer %s node has both roles and actors defined",
						destNode.ID,
					)
				}

				switch {
				case len(destNode.Roles) >= 1:
					if !destNode.RoleBroadcast || len(destNode.Roles) != 1 {
						panic("not implemented")
					}
					actors = p.actors.ListRole(ctx, api.Actor_CONSUMER, destNode.Roles[0])
				case len(destNode.Actors) >= 1:
					for _, actorRef := range destNode.Actors {
						actorID := api.MaybeUUIDFromString(actorRef)
						if !actorID.IsZero() {
							actor, exists := p.actors.Get(ctx, actorID)
							if !exists {
								return nil, false, fmt.Errorf("No such actor: %s", actorID)
							}
							actors = append(actors, actor)
						} else {
							actor, exists := p.actors.GetByName(ctx, actorRef)
							if !exists {
								return nil, false, fmt.Errorf("No such actor: %s", actorRef)
							}
							actors = append(actors, actor)
						}
					}
				default:
					panic("not implemented")
				}
			case storage.WorkerNode:
				if destNode.RoleBroadcast || len(destNode.Roles) != 0 {
					panic("Worker graph nodes supports only Actors, not roles")
				}

				for _, actorID := range destNode.ActorIDs {
					actor, exists := p.actors.Get(ctx, actorID)
					if !exists {
						return nil, false, fmt.Errorf("No such actor: %s", actorID)
					}
					actors = append(actors, actor)
				}

				for _, actorRef := range destNode.Actors {
					actorID := api.MaybeUUIDFromString(actorRef)
					if !actorID.IsZero() {
						actor, exists := p.actors.Get(ctx, actorID)
						if !exists {
							return nil, false, fmt.Errorf("No such actor: %s", actorID)
						}
						actors = append(actors, actor)
					} else {
						actor, exists := p.actors.GetByName(ctx, actorRef)
						if !exists {
							return nil, false, fmt.Errorf("No such actor: %s", actorRef)
						}
						actors = append(actors, actor)
					}
				}
			case storage.RequestorNode:
				actor, ok := p.actors.Get(ctx, p.TriggerEmitterID)
				if !ok {
					panic("The requestor disappeared during process. Please report")
				}
				actors = append(actors, actor)
			}

			if len(actors) == 0 {
				return nil, false, fmt.Errorf("No target actor found")
			}
			var targets []*api.EnvelopeTarget
			for _, actor := range actors {
				targets = append(targets, &api.EnvelopeTarget{
					ActorID: actor.ID,
					NodeID:  dest.NodeID,
					Input:   dest.Input,
				})
			}

			isResponse := destNode.Type == storage.RequestorNode && dest.Input == "response"
			return targets, isResponse, nil
		},
	)
}

// UpdateReceiveStatus is called whenever an envelope fragment is received from a node
func (p *Process) UpdateReceiveStatus(ctx context.Context, envelopeID, actorID api.UUID, outputref storage.NodeOutputRef, status storage.ReceptionStatus) error {
	return p.processStorage.UpdateReceiveStatus(
		ctx,
		storage.Reception{
			ProcessID:       p.ID,
			EnvelopeID:      envelopeID,
			SourceActor:     actorID,
			SourceOutputRef: outputref,
		},
		status)
}

// ActorProcessingEnd informs the process
func (p *Process) ActorProcessingEnd(ctx context.Context, actorID api.UUID, nodeID string, status api.ActorProcessingState_Status, log zerolog.Logger) {
	if err := p.processStorage.ProcessNodeEnd(
		ctx, p.ID, storage.ProcessNode{ActorID: actorID, NodeID: nodeID}, status,
	); err != nil {
		log.Err(err).Msg("Error calling ProcessNodeEnd")
	}
}

// GetResponse returns the process response envelope id if any
func (p *Process) GetResponse(ctx context.Context) api.UUID {
	state := p.processStorage.GetState(ctx, p.ID)
	return state.ResponseEnvelopeID
}

// GetPostMortemState set the postmortem state of the process
func (p *Process) GetPostMortemState(ctx context.Context) (storage.PostMortemState, error) {
	if status := p.GetStatus(ctx); status != api.Process_DONE && status != api.Process_ERROR {
		return storage.PostMortemState{}, fmt.Errorf("Cannot get postmortem state on a non-ended process. Current status is '%s'", status)
	}
	return p.processStorage.GetPostMortemState(ctx, p.ID), nil
}

// SetPostMortemState set the postmortem state of the process
func (p *Process) SetPostMortemState(ctx context.Context, state storage.PostMortemState) error {
	if status := p.GetStatus(ctx); status != api.Process_DONE && status != api.Process_ERROR {
		return fmt.Errorf("Cannot set postmortem state on a non-ended process. Current status is '%s'", status)
	}
	p.processStorage.SetPostMortemState(ctx, p.ID, state)
	return nil
}
