// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package process

import (
	"context"
	"sync"

	jsoniter "github.com/json-iterator/go"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/utils"

	"xbus.io/xbus/storage"
)

type ExportData struct {
	exportEnvelope    bool
	exportJobs        bool
	exportLogs        bool
	exportProcessLogs bool
	envelopeSizeLimit int64
	logLevel          api.LogLevel

	emitter     *api.Actor
	status      api.Process_Status
	pmState     *storage.PostMortemState
	state       storage.ProcessState
	jobs        []storage.Job
	logs        []*api.LogEntry
	processLogs []storage.ProcessLogEntry

	envelopeIDs   []api.UUID
	envelopeIDset map[api.UUID]bool
	envelopeStart map[api.UUID]*api.Envelope
	envelopePos   map[api.UUID]storage.EnvelopeReadPos
}

func (p *Process) LoadExportData(
	ctx context.Context,
	_ storage.EnvelopeStorage,
	jobStorage storage.JobStorage,
	logStorage storage.LogStorage,
	exportEnvelope bool,
	exportJobs bool,
	exportLogs bool,
	exportProcessLogs bool,
	envelopeSizeLimit int64,
	logLevel api.LogLevel,
	_ zerolog.Logger,
) (*ExportData, error) {
	var (
		loadWG sync.WaitGroup
		data   = ExportData{
			exportEnvelope:    exportEnvelope,
			exportJobs:        exportJobs,
			exportLogs:        exportLogs,
			exportProcessLogs: exportProcessLogs,
			envelopeSizeLimit: envelopeSizeLimit,
			logLevel:          logLevel,

			envelopeIDset: map[api.UUID]bool{},
			envelopeStart: map[api.UUID]*api.Envelope{},
			envelopePos:   map[api.UUID]storage.EnvelopeReadPos{},
		}
	)

	loadWG.Add(1)
	go func() {
		defer loadWG.Done()
		data.state = p.processStorage.GetState(ctx, p.ID)
		data.status = data.state.Status
		if data.status == api.Process_DONE || data.status == api.Process_ERROR {
			s, err := p.GetPostMortemState(ctx)
			if err != nil {
				panic(err)
			}
			data.pmState = &s
		}
	}()

	if exportJobs {
		loadWG.Add(1)
		go func() {
			defer loadWG.Done()
			data.jobs = jobStorage.ListJobByProcess(ctx, p.ID)
		}()
	}

	if exportLogs {
		loadWG.Add(1)
		go func() {
			defer loadWG.Done()
			data.logs = logStorage.GetProcessMessages(ctx, p.ID, logLevel)
		}()
	}

	if exportProcessLogs || exportEnvelope {
		loadWG.Add(1)
		go func() {
			defer loadWG.Done()
			data.processLogs = p.manager.ProcessStorage.GetProcessLogEntries(ctx, p.ID)

			if exportEnvelope {
				data.envelopeIDs = []api.UUID{p.TriggerEnvelopeID}
				data.envelopeIDset = map[api.UUID]bool{p.TriggerEnvelopeID: true}

				for _, log := range data.processLogs {
					if log.Emission != nil {
						if _, ok := data.envelopeIDset[log.Emission.EnvelopeID]; !ok {
							data.envelopeIDs = append(data.envelopeIDs, log.Emission.EnvelopeID)
							data.envelopeIDset[log.Emission.EnvelopeID] = true
						}
					}
				}
			}
		}()
	}

	loadWG.Wait()

	if actor, exists := p.actors.Get(ctx, p.TriggerEmitterID); exists {
		data.emitter = actor
	}

	return &data, nil
}

// ExportJSONStream exports a process data in a json stream
func (p *Process) ExportJSONStream(
	ctx context.Context,
	stream *jsoniter.Stream,
	envelopeStorage storage.EnvelopeStorage,
	jobStorage storage.JobStorage,
	logStorage storage.LogStorage,
	log zerolog.Logger,
	exportEnvelope bool,
	exportJobs bool,
	exportLogs bool,
	exportProcessLogs bool,
	envelopeSizeLimit int64,
	logLevel api.LogLevel,
) error {
	data, err := p.LoadExportData(ctx,
		envelopeStorage, jobStorage, logStorage,
		exportEnvelope, exportJobs, exportLogs, exportProcessLogs, envelopeSizeLimit, logLevel,
		log)
	if err != nil {
		return err
	}

	return p.ExportJSONStreamWithData(ctx, stream, envelopeStorage, data, log)
}

// ExportJSONStreamWithData exports a process data in a json stream
func (p *Process) ExportJSONStreamWithData(
	ctx context.Context,
	stream *jsoniter.Stream,
	envelopeStorage storage.EnvelopeStorage,
	data *ExportData,
	log zerolog.Logger,
) error {
	obj := utils.NewJSONStreamObject(stream)

	obj.String("id", p.ID.String())

	if !p.GroupID.IsZero() {
		obj.String("group_id", p.GroupID.String())
	}
	if !p.ReplayOf.IsZero() {
		obj.String("replay_of", p.ReplayOf.String())
	}

	obj.Time("created_at", p.CreatedAt)

	triggerObj := obj.Object("trigger")

	emitterObj := triggerObj.Object("emitter")
	emitterObj.String("id", p.TriggerEmitterID.String())
	if data.emitter != nil {
		emitterObj.String("name", data.emitter.Name)
	}
	emitterObj.End()

	triggerObj.String("envelope", p.TriggerEnvelopeID.String())
	triggerObj.End()

	if p.Pipeline != nil {
		obj.Val("pipeline", p.Pipeline)
	}

	obj.Val("status", data.status)

	if data.status == api.Process_DONE || data.status == api.Process_ERROR {
		obj.Val("postmortem_state", data.pmState)
	}

	obj.Val("state", data.state)

	if data.exportJobs {
		jobArr := obj.Array("jobs")
		for _, job := range data.jobs {
			jobArr.Val(job)
		}
		jobArr.End()
	}

	if data.exportLogs {
		logArr := obj.Array("logs")
		for _, log := range data.logs {
			logArr.Val(log)
		}
		logArr.End()
	}

	if data.exportProcessLogs {
		plogArr := obj.Array("processlogs")

		for _, log := range data.processLogs {
			plogArr.Val(log)
		}

		plogArr.End()
	}

	if data.exportEnvelope {
		envsObj := obj.Object("envelopes")

		var size int64

		for _, id := range data.envelopeIDs {
			if ctx.Err() != nil {
				return ctx.Err()
			}
			envelope, pos, err := envelopeStorage.ReadEnvelope(
				ctx, id, storage.EnvelopeStartPos(), envelopeExportChunkSize)
			if err != nil {
				log.Err(err).
					Str("envelope-id", id.String()).
					Msg("cannot read envelope")
				continue
			}
			size += envelopeExportChunkSize

			if pos.Complete {
				envsObj.Val(id.String(), envelope)
				continue
			}

			envArr := envsObj.Array(id.String())

			envArr.Val(envelope)

			lastPos := pos
			for !pos.Complete && (data.envelopeSizeLimit == 0 || size < data.envelopeSizeLimit) {
				envelope, pos, err = envelopeStorage.ReadEnvelope(
					ctx, id, lastPos, envelopeExportChunkSize,
				)
				if err != nil {
					log.Err(err).Str("envelope_id", id.String()).
						Msg("cannot read envelope")
					break
				}
				size += envelopeExportChunkSize
				if pos.Equals(lastPos) {
					log.Warn().Str("envelope-id", id.String()).
						Msg("envelope is not complete")
					break
				}
				envArr.Val(envelope)

				lastPos = pos
			}

			envArr.End()
		}

		envsObj.End()
	}

	obj.End()

	return nil
}
