// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package broker

import (
	"context"
	"fmt"

	nrpc "github.com/nats-rpc/nrpc"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/process"
	"xbus.io/xbus/storage"
)

func (s *Service) receiveEnvelope(
	ctx context.Context,
	actorID api.UUID,
	context *api.ProcessingContext,
	output string,
	envelope *api.Envelope,
	close bool,
) (
	process *process.Process,
	sourceRef storage.NodeOutputRef,
	ack api.EnvelopeAck,
	err error,
) {
	actor, exists := s.Opts.ActorStorage.Get(ctx, actorID)
	if !exists {
		err = api.ClientProtoError("Unknown actor: %s", actorID)
		return
	}

	// If no context is given, make sure the sender is as emitter
	if (context == nil || context.GetProcessIDAsUUID().IsZero()) &&
		actor.Kind != api.Actor_EMITTER {
		msg := fmt.Sprintf("Context-less envelope sent by non-emitter %s", actor.Name)
		s.Log.Error().Msg(msg)
		err = api.ClientProtoError(msg)
		return
	}

	// A context is given, make sure the process exists
	if context != nil && !context.GetProcessIDAsUUID().IsZero() {
		process, err = s.processManager.Get(ctx, context.GetProcessIDAsUUID())
		if err != nil {
			ack = *api.NewEnvelopeAckError(ack.GetIDAsUUID(), err)
			err = nil
			return
		}
		// TODO make sure NodeID matches the sending actor
		sourceRef.NodeID = context.NodeID
		sourceRef.Output = output
	}

	if close {
		ack.Status = api.EnvelopeAck_ACCEPTED

		err = process.UpdateReceiveStatus(
			ctx, api.UUID{}, actor.GetIDAsUUID(), sourceRef, storage.ReceptionClosed)
		return
	}

	if envelope != nil {
		s.Log.Debug().
			Str("envelope-id", envelope.GetIDAsUUID().String()).
			Msg("Receiving envelope fragment")
		ack.ID = envelope.ID
	} else {
		err = api.ClientProtoError("Missing envelope")
		s.Log.Err(err).Msg("")
		return
	}

	var (
		typesKnown bool
		eventTypes []string
	)

	// Check for empty envelope
	if envelope.Last && len(envelope.Events) == 0 && len(envelope.EventIDs) == 0 {
		if process == nil {
			ack = *api.NewEnvelopeAckError(
				envelope.GetIDAsUUID(),
				fmt.Errorf("Invalid envelop (missing or invalid context)"))
			return
		}

		// empty envelope, which can mean 2 things depending on the ID

		// Without an ID: the envelope is just a signal that the output
		// will not be used in this context. Inform the process.
		if envelope.GetIDAsUUID().IsZero() {
			panic("Closing is not done here, should have done already")
		}
		// With an ID: the envelope is one sent on an input, hence must be
		// known to the broker which can reread it from storage.
		ack.Status = api.EnvelopeAck_ACCEPTED
		typesKnown = true
	} else {
		// non-empty envelope. Store it.
		var status storage.EnvelopeState

		for _, event := range envelope.Events {
			if t := event.GetType(); t != "" && !tools.ValidateEventType(t) {
				ack.Status = api.EnvelopeAck_ERROR
				ack.Reason = fmt.Sprintf("Invalid event type: '%s'", event.GetType())
				return
			}
		}

		status, err = s.Opts.EnvelopeStorage.StoreEnvelope(ctx, envelope)
		if err != nil {
			panic(err)
		}
		s.Log.Debug().
			Str("envelope-id", envelope.GetIDAsUUID().String()).
			Str("status", status.Status.String()).
			Msg("Stored envelope")
		switch status.Status {
		case storage.EnvelopeStorageReceptionReceiving:
			ack.Status = api.EnvelopeAck_RECEIVING
		case storage.EnvelopeStorageReceptionComplete:
			ack.Status = api.EnvelopeAck_ACCEPTED
		case storage.EnvelopeStorageReceptionError:
			ack.Status = api.EnvelopeAck_ERROR
		default:
			ack.Status = api.EnvelopeAck_ERROR
		}
		typesKnown = status.TypesKnown
		eventTypes = status.EventTypes()

		if ack.Status == api.EnvelopeAck_ERROR {
			if process != nil {
				err = process.UpdateReceiveStatus(
					ctx, envelope.GetIDAsUUID(), actor.GetIDAsUUID(), sourceRef, storage.ReceptionError)
			}
			ack = *api.NewEnvelopeAckError(envelope.GetIDAsUUID(), status)
			return
		}
	}

	// The envelope is now safely stored
	// We just need the process to be aware of this reception, and we will
	// be done.

	if context == nil || context.GetProcessIDAsUUID().IsZero() {
		process, err = s.processManager.LookupOrCreate(
			ctx, actor, output, envelope.GetIDAsUUID())
		if err != nil {
			panic(fmt.Errorf("error getting or creating process: %w", err))
		}
		if process.Pipeline != nil {
			sourceRef = process.SourceRef
		}
	}

	// No context but all types known -> the process can match a pipeline if needed
	if process.Pipeline == nil && (context == nil || context.GetProcessIDAsUUID().IsZero()) && typesKnown {
		if matchErr := process.MatchPipeline(ctx, output, eventTypes); matchErr != nil {
			ack.Status = api.EnvelopeAck_ERROR
			ack.Reason = matchErr.Error()
			return
		}
		sourceRef = process.SourceRef
	}

	// If process is unknown and the reception is over, we have a serious
	// problem
	if process == nil && ack.Status == api.EnvelopeAck_ACCEPTED {
		panic("Impossible state, please report")
	}

	if process != nil && process.Pipeline != nil {
		// If a process is associated to the envelope, make sure the route for the next
		// move is known, then update the reception status.
		// Logically we should calculate the route after updating the status, but if the
		// server crashes we would have route-less receptions that would be hard to
		// track and fix. Switching the two operations is harmless: a reception-less
		// route will become used as soon as the emitter re-attempt its sending.
		// And because the route is present, the emission service will find what
		// to do as soon as it is (re)started
		// See issue #39

		_, err := process.GetTargets(ctx, sourceRef, envelope.GetIDAsUUID(), false)
		if err != nil {
			panic(fmt.Sprintf("Error getting routes: %s", err))
		}

		var storageReceptionStatus storage.ReceptionStatus
		switch ack.Status {
		case api.EnvelopeAck_RECEIVING:
			storageReceptionStatus = storage.ReceptionOngoing
		case api.EnvelopeAck_ACCEPTED:
			storageReceptionStatus = storage.ReceptionDone
		}
		if err := process.UpdateReceiveStatus(
			ctx, envelope.GetIDAsUUID(), actor.GetIDAsUUID(), sourceRef, storageReceptionStatus,
		); err != nil {
			panic(fmt.Sprintf("Error updating receive status: %s", err))
		}
	}

	// our part (receive the envelope fragment) is done, we can safely
	// let the ack be sent to the sender

	return
}

type brokerServer struct {
	*Service
}

// Output implements the Actor.Output api
func (s brokerServer) Output(
	ctx context.Context,
	req *api.OutputRequest,
) (
	*api.EnvelopeAck, error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	actorSID := nrpc.GetRequest(ctx).ServiceParam("actorID")
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return nil, err
	}

	_, _, ack, err := s.receiveEnvelope(
		ctx, actorID, req.Context, req.Output, req.Envelope, req.Close)

	return &ack, err
}
