// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package control

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
)

type actorServer struct {
	*Service
}

func (s actorServer) Create(
	ctx context.Context, req *control.ActorRequest,
) (*control.ActorReply, error) {
	for _, actor := range req.ActorList {
		if !actor.GetIDAsUUID().IsZero() {
			return nil, fmt.Errorf("bad request: actor.id must be unset")
		}
		id, err := api.UUIDV4()
		if err != nil {
			return nil, fmt.Errorf("could not create a uuid: %w", err)
		}
		actor.ID = id.ToBytes()
	}

	var resp control.ActorReply
	for _, actor := range req.ActorList {
		sActor, err := s.o.ActorStorage.Create(ctx, actor)
		if err != nil {
			return nil, err
		}
		actor = &api.Actor{
			ID:     sActor.ID,
			Name:   sActor.Name,
			Kind:   sActor.Kind,
			Status: sActor.Status,
			Roles:  sActor.Roles,
		}

		resp.ActorList = append(resp.ActorList, actor)
	}

	return &resp, nil
}

func (s actorServer) Update(ctx context.Context, req *control.ActorRequest) (resp *control.ActorReply, err error) {
	resp = &control.ActorReply{}
	for _, actor := range req.ActorList {
		sActor, exist := s.o.ActorStorage.Get(ctx, actor.GetIDAsUUID())
		if !exist {
			err = fmt.Errorf("No such actor: %s", actor.GetIDAsUUID())
			return
		}

		switch {
		case actor.Status == api.Actor_ENABLED && sActor.Status != api.Actor_ENABLED:
			sActor, err = s.actorAccept(ctx, actor.GetIDAsUUID())
		case actor.Status == api.Actor_REJECTED:
			sActor, err = s.actorReject(ctx, actor.GetIDAsUUID())
		case actor.Status == api.Actor_ENABLED:
			sActor.Config = actor.Config
			sActor, err = s.o.ActorStorage.Update(ctx, sActor)
			if err != nil {
				return
			}
		default:
			err = api.ClientProtoError("Invalid actor update")
			return
		}
		actor = &api.Actor{
			ID:     sActor.ID,
			Name:   sActor.Name,
			Kind:   sActor.Kind,
			Status: sActor.Status,
			Roles:  sActor.Roles,
		}

		if err != nil {
			break
		}
		resp.ActorList = append(resp.ActorList, actor)
	}
	return
}

// Delete deletes an Actor
func (s actorServer) Delete(ctx context.Context, req *control.ActorRequest) (resp *control.ActorReply, err error) {
	for _, actor := range req.ActorList {
		s.o.ActorStorage.Delete(ctx, actor.GetIDAsUUID())
	}
	return
}

// Query lis Actors
func (s actorServer) Query(
	ctx context.Context, req *control.ActorRequest,
) (*control.ActorReply, error) {
	list, err := s.actorQuery(ctx, req.GetActorList())
	if err != nil {
		return nil, err
	}
	return &control.ActorReply{
		ActorList: list,
	}, nil
}

func (s *Service) actorQuery(ctx context.Context, filters []*api.Actor) (res []*api.Actor, err error) {
	if len(filters) == 0 {
		return s.o.ActorStorage.List(ctx), nil
	}
	for _, filter := range filters {
		switch {
		case !filter.GetIDAsUUID().IsZero():
			if actor, exis := s.o.ActorStorage.Get(ctx, filter.GetIDAsUUID()); exis {
				res = append(res, actor)
			}
		case filter.Name != "":
			if actor, exis := s.o.ActorStorage.GetByName(ctx, filter.Name); exis {
				res = append(res, actor)
			}
		case !filter.GetAccountIDAsUUID().IsZero():
			res = append(res, s.o.ActorStorage.ListAccount(ctx, filter.GetAccountIDAsUUID())...)
		default:
			panic("invalid filter")
		}
	}
	return
}

func (s *Service) actorAccept(ctx context.Context, id api.UUID) (*api.Actor, error) {
	a, exis := s.o.ActorStorage.Get(ctx, id)
	if !exis {
		return nil, fmt.Errorf("No such actor: %s", id)
	}
	if a.Status != api.Actor_PENDING {
		return nil, fmt.Errorf("Invalid initial status '%s' for 'Accept'", a.Status)
	}
	a.Status = api.Actor_ENABLED
	return s.o.ActorStorage.Update(ctx, a)
}

func (s *Service) actorReject(ctx context.Context, id api.UUID) (*api.Actor, error) {
	a, exis := s.o.ActorStorage.Get(ctx, id)
	if !exis {
		return nil, fmt.Errorf("No such actor: %s", id)
	}
	if a.Status != api.Actor_PENDING {
		return nil, fmt.Errorf("Invalid initial status '%s' for 'Reject'", a.Status)
	}
	a.Status = api.Actor_REJECTED
	return s.o.ActorStorage.Update(ctx, a)
}
