// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package control

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	math_bits "math/bits"
	"sync"
	"sync/atomic"

	jsoniter "github.com/json-iterator/go"
	"google.golang.org/protobuf/proto"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/process"
)

var processExportJSONConfig = jsoniter.Config{
	EscapeHTML:                    false,
	IndentionStep:                 2,
	ObjectFieldMustBeSimpleString: true, // do not unescape object field
	ValidateJsonRawMessage:        true,
}.Froze()

type processServer struct {
	*Service
}

func (s processServer) Query(
	ctx context.Context, req *control.ProcessQueryRequest,
	send func(*control.ProcessQueryReply),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	pList := s.o.ProcessStorage.Query(ctx, req.Filter)
	select {
	case <-ctx.Done():
		return nil
	default:
	}
	if len(pList) == 0 {
		return nil
	}
	var (
		maxSize = 1024 * 512
		size    = 0
		batch   = make([]*api.Process, 0, 256)
	)
	for _, p := range pList {
		state := s.o.ProcessStorage.GetState(ctx, p.ID)
		b, err := json.Marshal(state)
		if err != nil {
			panic(err)
		}
		createdAt := p.CreatedAt
		entry := api.Process{
			ID:                p.ID.ToBytes(),
			TriggerEmitterID:  p.TriggerEmitterID.ToBytes(),
			TriggerEnvelopeID: p.TriggerEnvelopeID.ToBytes(),
			CreatedAt:         timestamppb.New(createdAt),
			PipelineID:        p.PipelineID.ToBytes(),
			GroupID:           p.GroupID.ToBytes(),
			ReplayOf:          p.ReplayOf.ToBytes(),
			Status:            state.Status,
			State:             string(b),
		}
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		entrySize := proto.Size(&entry)
		entrySizeDelta := entrySize + 1 + sovControl(uint64(entrySize))
		if size+entrySizeDelta > maxSize {
			send(&control.ProcessQueryReply{ProcessList: batch})
			batch = batch[0:0]
			size = 0
		}
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		batch = append(batch, &entry)
		size += entrySizeDelta
	}
	if len(batch) != 0 {
		send(&control.ProcessQueryReply{ProcessList: batch})
	}
	return nil
}

// sovControl is imported from control.pb.go and is used to calculate the
// ProcessSummary size, that we need to evaluate while building the entries
// list
func sovControl(x uint64) (n int) {
	return (math_bits.Len64(x|1) + 6) / 7
}

func (s processServer) QuerySummary(
	ctx context.Context, req *control.ProcessQueryRequest,
	send func(*control.ProcessSummary),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	summary, err := s.o.ProcessStorage.QuerySummary(ctx, req.Filter)
	if err != nil {
		return err
	}
	if len(summary) == 0 {
		return nil
	}
	var (
		entries = make([]*control.ProcessSummary_Entry, 0, len(summary))
		maxSize = 1024 * 512
		size    = 0
	)
	for _, entry := range summary {
		entry := control.ProcessSummary_Entry{
			PipelineID:  entry.PipelineID.ToBytes(),
			EmitterID:   entry.EmitterID.ToBytes(),
			Status:      entry.Status,
			ResultAcked: entry.ResultAcked,
			Count:       entry.Count,
		}
		entrySize := proto.Size(&entry)
		entrySizeDelta := entrySize + 1 + sovControl(uint64(entrySize))
		if size+entrySizeDelta > maxSize {
			send(&control.ProcessSummary{Entries: entries})
			entries = entries[0:0]
			size = 0
		}
		entries = append(entries, &entry)
		size += entrySizeDelta
	}
	if len(entries) != 0 {
		send(&control.ProcessSummary{Entries: entries})
	}

	return nil
}

func (s processServer) Control(
	ctx context.Context, req *control.ProcessControlRequest,
) (err error) {
	s.TaskAdd()
	defer s.TaskDone()
	p, err := s.o.ProcessManager.Get(ctx, req.GetProcessIDAsUUID())
	if err != nil {
		return err
	}
	switch req.Command {
	case control.ProcessControlRequest_PAUSE:
		err = p.Pause(ctx)
	case control.ProcessControlRequest_RESUME:
		err = p.Resume(ctx)
	case control.ProcessControlRequest_CANCEL:
		err = p.Cancel(ctx, "Canceled by user")
	default:
		err = fmt.Errorf("invalid command: %w", err)
	}
	return
}

func (s processServer) Export(
	_ context.Context, _ *control.ProcessExportRequest,
) (
	resp *control.ProcessExportReply, err error,
) {
	return nil, fmt.Errorf("'Process.Export' is no longer available. Please use 'Process.ExportStream'")
}

func (s processServer) ExportStream(
	ctx context.Context, req *control.ProcessExportRequest,
	send func(*control.OctetStreamList),
) (
	err error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	// Encoded data pass through this pipe to be chunked and sent
	r, w := io.Pipe()
	defer w.Close()

	// Receives the process exported data
	type exportDataJob struct {
		p          *process.Process
		exportData *process.ExportData
		err        error
	}
	exportDataCh := make(chan *exportDataJob, 1)

	var wg sync.WaitGroup
	wg.Add(1)

	var currentIndex int64

	// Chunk and send the encoded data
	go func() {
		defer wg.Done()
		var (
			bufSize = 1024 * 768
			minSize = 1024 * 512
			pos     int
		)

		buffer := make([]byte, bufSize)

		var chunkIndex int64
		for {
			if ctx.Err() != nil {
				return
			}
			n, err := r.Read(buffer[pos:])
			if ctx.Err() != nil {
				return
			}
			pos += n
			if pos > 0 && (pos >= minSize || errors.Is(err, io.EOF)) {
				index := atomic.LoadInt64(&currentIndex)
				if ctx.Err() == nil {
					send(&control.OctetStreamList{
						Progression:    index,
						MaxProgression: int64(len(req.ProcessIDs)),
						Chunks: []*control.OctetStreamList_Chunk{
							{
								Data:  buffer[:pos],
								Index: chunkIndex,
							},
						},
					})
					chunkIndex++
				}
				pos = 0
			}
			if errors.Is(err, io.EOF) {
				return
			}
		}
	}()

	// Export the process data
	go func() {
		defer close(exportDataCh)
		for _, pid := range req.GetProcessIDsAsUUIDSlice() {
			if ctx.Err() != nil {
				return
			}
			p, err := s.o.ProcessManager.Get(ctx, pid)
			if err != nil {
				exportDataCh <- &exportDataJob{
					err: err,
				}
				break
			}
			data, err := p.LoadExportData(
				ctx,
				s.o.EnvelopeStorage,
				s.o.JobStorage,
				s.o.LogStorage,
				req.ExportFull || req.ExportEnvelope,
				req.ExportFull || req.ExportJobs,
				req.ExportFull || req.ExportLogs,
				req.ExportFull || req.ExportProcessLogs,
				req.EnvelopeSizeLimit,
				req.LogLevel,
				s.Log,
			)
			if err != nil {
				exportDataCh <- &exportDataJob{
					err: err,
				}
				break
			}
			exportDataCh <- &exportDataJob{p: p, exportData: data}
		}
	}()

	stream := processExportJSONConfig.BorrowStream(w)
	defer processExportJSONConfig.ReturnStream(stream)

	stream.WriteArrayStart()

	for job := range exportDataCh {
		if job.err != nil {
			return err
		}
		index := atomic.AddInt64(&currentIndex, 1)
		if index != 1 {
			stream.WriteMore()
		}
		if ctx.Err() != nil {
			return ctx.Err()
		}
		if err := job.p.ExportJSONStreamWithData(
			ctx,
			stream,
			s.o.EnvelopeStorage,
			job.exportData,
			s.Log,
		); err != nil {
			panic(err)
		}
		if err := stream.Flush(); err != nil {
			return err
		}
	}

	stream.WriteArrayEnd()

	if err := stream.Flush(); err != nil {
		return err
	}

	if err := w.Close(); err != nil {
		return err
	}
	wg.Wait()
	return
}

func (s processServer) Purge(
	ctx context.Context, req *control.ProcessPurgeRequest,
	send func(*control.TaskProgress),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	if len(req.ProcessIDs) == 0 {
		send(&control.TaskProgress{})
		return nil
	}

	jobCount := len(req.ProcessIDs) / s.o.PurgeBatchSize
	if jobCount == 0 {
		jobCount = 1
	} else if jobCount > s.o.MaxPurgeJobs {
		jobCount = s.o.MaxPurgeJobs
	}

	type res struct {
		err error
		ids []api.UUID
	}

	jobsCh := make(chan []api.UUID)
	resCh := make(chan res, jobCount)
	stopCh := make(chan struct{})

	var jobGroup sync.WaitGroup

	// Spin up jobs
	for i := 0; i != jobCount; i++ {
		jobGroup.Add(1)
		go func() {
			defer jobGroup.Done()
			for {
				select {
				case batch, ok := <-jobsCh:
					if !ok {
						return
					}
					ids, err := s.o.ProcessManager.PurgeMany(ctx, batch)
					if err == nil {
						err = s.o.JobStorage.Purge(ctx, ids)
					}
					resCh <- res{err, ids}
				case <-ctx.Done():
					return
				}
			}
		}()
	}

	var wg sync.WaitGroup
	defer wg.Wait()
	wg.Add(1)
	go func() {
		defer wg.Done()
		defer close(resCh)
		jobGroup.Wait()
	}()

	var (
		max     = uint32(len(req.ProcessIDs))
		counter uint32
	)

	wg.Add(1)
	// Push jobs to the queue
	go func() {
		defer wg.Done()
		defer close(jobsCh)
		var batches [][]api.UUID

		ids := req.GetProcessIDsAsUUIDSlice()
		for s.o.PurgeBatchSize < len(ids) {
			ids, batches = ids[s.o.PurgeBatchSize:], append(batches, ids[0:s.o.PurgeBatchSize:s.o.PurgeBatchSize])
		}
		batches = append(batches, ids)
		for _, batch := range batches {
			select {
			case <-stopCh:
				return
			case <-ctx.Done():
				return
			case jobsCh <- batch:
			}
		}
	}()

	// Receive results and report back to the caller
	for {
		select {
		case <-ctx.Done():
			return nil

		case r, ok := <-resCh:
			if !ok {
				return nil
			}
			if r.err != nil {
				stopCh <- struct{}{}
				return r.err
			}
			counter += uint32(len(r.ids))
			send(&control.TaskProgress{Progression: counter, MaxProgression: max})
		}
	}
}

func (s processServer) Replay(
	ctx context.Context, req *control.ProcessReplayRequest,
) (*api.Process, error) {
	process, err := s.o.ProcessManager.Get(ctx, req.Process.GetIDAsUUID())
	if err != nil {
		return nil, fmt.Errorf("could not get process: %w", err)
	}
	newProcess, err := s.o.ProcessManager.Replay(ctx, process, s.o.EnvelopeStorage, req.Match)
	if err != nil {
		return nil, fmt.Errorf("could not replay process: %w", err)
	}
	state := s.o.ProcessStorage.GetState(ctx, newProcess.ID)
	b, err := json.Marshal(state)
	if err != nil {
		panic(err)
	}
	createdAt := newProcess.CreatedAt
	return &api.Process{
		ID:                newProcess.ID.ToBytes(),
		TriggerEmitterID:  newProcess.TriggerEmitterID.ToBytes(),
		TriggerEnvelopeID: newProcess.TriggerEnvelopeID.ToBytes(),
		CreatedAt:         timestamppb.New(createdAt),
		PipelineID:        newProcess.Pipeline.ID,
		GroupID:           newProcess.GroupID.ToBytes(),
		ReplayOf:          newProcess.ReplayOf.ToBytes(),
		Status:            state.Status,
		State:             string(b),
	}, nil
}
