// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package core

import (
	"context"
	"testing"

	_ "github.com/orus-io/qb/dialects/postgres"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql"
	"xbus.io/xbus/testutils"
	"xbus.io/xbus/testutils/sqldb"
)

func TestUpdateState(t *testing.T) {
	ctx := testutils.GetTestContext(t)
	logger := testutils.GetTestLogger(t)
	zlog := *logger.Zero()

	testdb := sqldb.GetTestDB(t)
	defer testdb.Close()
	model, db := testdb.GetModelDB(1)

	actorStorage := sql.NewActorStorage(model, db, zlog)
	pipelineStorage := sql.NewPipelineStorage(model, db, zlog)
	pipelineManager := pipeline.NewManager(actorStorage, pipelineStorage)
	processStorage := sql.NewProcessStorage(model, db, zlog)
	envelopeStorage := sql.NewEnvelopeStorage(model, db, zlog)

	pipelineID, err := pipelineStorage.Save(ctx, storage.Pipeline{
		PipelineInfo: &api.PipelineInfo{
			Name:    "default",
			Version: "0.1",
		},
		Graph: storage.Graph{
			Nodes: []storage.Node{
				{
					ID:            "source",
					Type:          storage.EmitterNode,
					Roles:         []string{"default"},
					RoleBroadcast: true,
				},
				{
					ID:            "sink",
					Type:          storage.ConsumerNode,
					Roles:         []string{"default"},
					RoleBroadcast: true,
				},
			},
			Edges: []storage.Edge{
				{
					From: storage.NodeOutputRef{NodeID: "source", Output: "default"},
					To:   storage.NodeInputRef{NodeID: "sink", Input: "default"},
				},
			},
		},
	})
	if err != nil {
		t.Fatal(err)
	}
	if err := pipelineStorage.SetStatus(ctx, pipelineID, api.PipelineInfo_ACTIVE); err != nil {
		t.Fatal(err)
	}

	emitter, err := actorStorage.Create(ctx, &api.Actor{
		Name:  "emitter",
		Kind:  api.Actor_EMITTER,
		Roles: []string{"default"},
	})
	if err != nil {
		t.Fatal(err)
	}
	emID := emitter.GetIDAsUUID()

	envID := api.MustUUIDFromString("2dd179e8-b40b-11e7-9aad-7fcecdf65470")
	evID := api.MustUUIDFromString("df854950-b40e-11e7-a515-bb760274e30c")

	if _, err := envelopeStorage.StoreEnvelope(ctx, &api.Envelope{
		ID:       envID.ToBytes(),
		EventIDs: api.UUIDSlice{evID}.ToBytes(),
		Events: []*api.Event{
			{
				ID:        evID.ToBytes(),
				Items:     [][]byte{},
				ItemCount: 0,
				Checksum:  0,
			},
		},
		Last: true,
	}); err != nil {
		t.Fatal(err)
	}

	service := Options{
		ActorStorage:    actorStorage,
		ProcessStorage:  processStorage,
		ProcessManager:  process.NewManager(pipelineManager, actorStorage, processStorage, *logger.Zero()),
		PipelineManager: pipelineManager,
		LogStorage:      sql.NewLogStorage(model, db, zlog),
		Options: services.Options{
			Logger: *logger.Zero(),
		},
	}.NewService(context.TODO())

	require.NoError(t, service.BaseService.Startup())

	var logentries []storage.ProcessLogEntry

	processStorage.SetProcessLogEntryHandler(func(entries []storage.ProcessLogEntry) {
		logentries = append(logentries, entries...)
	})

	pID, err := processStorage.CreateProcess(ctx, storage.Process{
		TriggerEmitterID:  emID,
		TriggerEnvelopeID: envID,
		PipelineID:        pipelineID,
		SourceOutput:      storage.NodeOutputRef{NodeID: "source", Output: "default"},
	})
	if err != nil {
		t.Fatal(err)
	}

	if err := processStorage.UpdateReceiveStatus(
		ctx,
		storage.Reception{
			ProcessID:   pID,
			EnvelopeID:  envID,
			SourceActor: emID,
			SourceOutputRef: storage.NodeOutputRef{
				NodeID: "source",
				Output: "default",
			},
		},
		storage.ReceptionDone,
	); err != nil {
		t.Fatal(err)
	}

	if err := processStorage.ProcessNodeEnd(
		ctx,
		pID, storage.ProcessNode{ActorID: emID, NodeID: "sink"},
		api.ActorProcessingState_SUCCESS,
	); err != nil {
		t.Fatal(err)
	}

	logger.ExpectsError(testutils.ExpectsContains("Error calculating next state: "))
	for _, entry := range logentries {
		service.updateState(entry.ProcessID, []storage.ProcessLogEntry{entry})
	}

	state := processStorage.GetState(ctx, pID)
	assert.Equal(t, api.Process_ERROR, state.Status)
	assert.Contains(t, state.StatusReason, "Could not compute next state: Error calculating next state:")

	var count int
	if err := db.Query(model.ProcessLogEntry).Where(
		model.ProcessLogEntry.LogProcessed.Eq(false),
	).Count(&count); err != nil {
		t.Fatal(err)
	}
	assert.Equal(t, 0, count)

	if err := db.Query(model.ProcessLogEntry).Where(
		model.ProcessLogEntry.LogProcessed.Eq(true),
		model.ProcessLogEntry.ProcessingError.Eq(false),
	).Count(&count); err != nil {
		t.Fatal(err)
	}
	assert.Equal(t, 1, count)

	if err := db.Query(model.ProcessLogEntry).Where(
		model.ProcessLogEntry.LogProcessed.Eq(true),
		model.ProcessLogEntry.ProcessingError.Eq(true),
	).Count(&count); err != nil {
		t.Fatal(err)
	}
	assert.Equal(t, 1, count)
}
