// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package director

import (
	"context"
	"fmt"

	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
)

type directorServer struct {
	*Service
}

func (s directorServer) Log(
	ctx context.Context, actorSID string, req *api.ActorLogRequest,
) (err error) {
	fmt.Println("CONTEXT", ctx)
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	if req.EnvelopeID != nil && !req.GetEnvelopeIDAsUUID().IsZero() {
		_, err = s.Opts.ProcessManager.Get(ctx, req.Context.GetProcessIDAsUUID())
		if err != nil {
			return
		}
	}

	var (
		envID, processID api.UUID
		nodeID           string
	)
	if req.EnvelopeID != nil {
		envID = req.GetEnvelopeIDAsUUID()
	}
	if ctx := req.Context; ctx != nil {
		nodeID = ctx.NodeID
		processID = ctx.GetProcessIDAsUUID()
	}
	for _, msg := range req.Messages {
		s.Opts.Logger.Info().Interface("msg", msg).Msg("logging")
		s.Opts.LogStorage.Log(ctx, &api.LogEntry{
			EnvelopeID: envID.ToBytes(),
			ActorID:    actorID.ToBytes(),
			NodeID:     nodeID,
			ProcessID:  processID.ToBytes(),
			Message:    msg,
		})
	}
	return
}

func (s directorServer) ProcessingEnd(
	ctx context.Context, actorSID string, req *api.ActorProcessingState,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	p, err := s.Opts.ProcessManager.Get(ctx, req.Context.GetProcessIDAsUUID())
	if err != nil {
		return
	}

	if req.Status != api.ActorProcessingState_SUCCESS &&
		req.Status != api.ActorProcessingState_ERROR {
		return fmt.Errorf("Invalid status: %s", req.Status)
	}

	for _, msg := range req.Messages {
		s.Opts.LogStorage.Log(ctx, &api.LogEntry{
			ActorID:   actorID.ToBytes(),
			NodeID:    req.Context.NodeID,
			ProcessID: req.Context.ProcessID,
			Message:   msg,
		})
	}

	s.Opts.JobStorage.SetState(
		ctx,
		storage.JobID{
			ProcessID: req.Context.GetProcessIDAsUUID(),
			Node: storage.ProcessNode{
				ActorID: actorID,
				NodeID:  req.Context.NodeID,
			},
		}, storage.JobState{
			Detached: true,
			Status:   req.Status,
		})
	p.ActorProcessingEnd(ctx, actorID, req.Context.NodeID, req.Status, s.Opts.Logger)
	return
}

func (s directorServer) GetRunningActorRequestList(
	ctx context.Context, actorSID string,
	send func(*api.ActorProcessRequest),
) error {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return err
	}
	jobList, err := s.Opts.JobStorage.ListDetachedJobByActor(ctx, actorID)
	if err != nil {
		return err
	}
	defer jobList.Close()
	for jobList.Next() && ctx.Err() == nil {
		job, err := jobList.Get()
		if err != nil {
			return err
		}
		req := api.ActorProcessRequest{
			Context: &api.ProcessingContext{
				ProcessID: job.ProcessID.ToBytes(),
				NodeID:    job.Node.NodeID,
			},
		}

		inputs, err := s.processManager.ProcessStorage.GetInputs(
			ctx, job.ProcessID, job.Node,
		)
		if err != nil {
			return err
		}
		for _, input := range inputs {
			envelope, pos, err := s.Opts.EnvelopeStorage.ReadEnvelope(
				ctx, input.EnvelopeID, storage.EnvelopeStartPos(), 100*1024)
			if err != nil {
				s.Service.Log.Error().
					Str("envelope-id", input.EnvelopeID.String()).
					Msg("could not read envelope")
				return err
			}
			req.Inputs = append(req.Inputs, &api.ActorProcessRequest_Input{
				Name:     input.Name,
				Envelope: envelope,
				Position: storage.EncodePosition(envelope.GetIDAsUUID(), pos),
			},
			)
		}
		if ctx.Err() == nil {
			send(&req)
		}
	}
	return ctx.Err()
}
