// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package envelopestorage

import (
	"context"
	"fmt"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "envelopestorage"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// EnvelopeStallDelay is the delay after which an envelope is marked stalled
// when its emitter stops sending fragments.
// TODO in a later version, this value should be envelope-dependant and be
// set by the emitter itself (within reasonnable limits)
var EnvelopeStallDelay = 30 * time.Second

// Options is used to create a 'envelopestorage' service
type Options struct {
	services.Options

	EnvelopeStorage storage.EnvelopeStorage
	JobStorage      storage.JobStorage
	ProcessStorage  storage.ProcessStorage
}

// NewService creates a Service
func (o Options) NewService(ctx context.Context) *Service {
	return &Service{
		BaseService: services.NewBaseService(ctx, ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	workers *nrpc.WorkerPool

	o Options
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	// TODO make the pool settings changeable by the xbusd configuration
	s.workers = nrpc.NewWorkerPool(s.Ctx, 100, 100, 1*time.Minute)

	envelopeStorageHandler := api.NewEnvelopeStorageConcurrentHandler(
		s.workers, s.Conn, envelopeStorageServer{s},
	)
	return s.SubscribeAll(
		s.HandlerQueueSubscribe(envelopeStorageHandler, QueueName),
	)
}

// Shutdown stops the service
func (s *Service) Shutdown() error {
	err := s.BaseService.Shutdown()
	s.workers.Close(time.Second)
	return err
}

type envelopeStorageServer struct {
	*Service
}

func (s envelopeStorageServer) Retrieve(
	ctx context.Context, clientKind string, clientID string,
	req *api.EnvelopePosition, send func(*api.EnvelopeFragmentAndPosition),
) error {
	// TODO verify credentials
	if clientKind == "actor" {
		actorID, err := api.UUIDFromString(clientID)
		if err != nil {
			return err
		}
		allowed := false
		for _, job := range s.o.JobStorage.ListRunningJobByEnvelopeID(ctx, req.GetEnvelopeIDAsUUID()) {
			if job.Node.ActorID == actorID {
				allowed = true
			}
		}
		if !allowed {
			isResponse, err := s.o.ProcessStorage.EnvelopeIsResponseFor(ctx, req.GetEnvelopeIDAsUUID(), actorID)
			if err != nil {
				return err
			}
			if !isResponse {
				return fmt.Errorf("Permission denied")
			}
		}
	}
	pos := storage.DecodePosition(req)
	for !pos.Complete {
		if ctx.Err() != nil {
			return ctx.Err()
		}
		fragment, newpos, err := s.o.EnvelopeStorage.ReadEnvelope(
			ctx, req.GetEnvelopeIDAsUUID(), pos, 1024*512)
		if err != nil {
			s.Log.Err(err).
				Str("envelope-id", req.GetEnvelopeIDAsUUID().String()).
				Msg("failed to read envelope")
			panic(err)
		}
		// TODO add a blocking version of readenvelope that waits until
		// there is something to return
		if newpos.Equals(pos) {
			s.o.EnvelopeStorage.CheckStalled(ctx, req.GetEnvelopeIDAsUUID(), EnvelopeStallDelay)
			time.Sleep(250 * time.Millisecond)
		} else {
			pos = newpos
			send(&api.EnvelopeFragmentAndPosition{
				Fragment: fragment,
				Position: storage.EncodePosition(req.GetEnvelopeIDAsUUID(), pos),
			})
		}
	}
	return nil
}
