// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"
	"errors"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage/sql/model"
)

// NewAccountStorage initialize a new AccountStorage
func NewAccountStorage(model *model.Model, db *yago.DB, logger zerolog.Logger) *AccountStorage {
	return &AccountStorage{Base{
		model:  model,
		db:     db,
		logger: logger,
	}}
}

// AccountStorage implements all the persistor interfaces
type AccountStorage struct {
	Base
}

func toDB(sAccount *api.Account) model.Account {
	return model.Account{
		Base:      model.Base{ID: uuid.UUID(sAccount.GetIDAsUUID())},
		Name:      sAccount.Name,
		Type:      sAccount.Type,
		Status:    sAccount.Status,
		CSR:       sAccount.CSR,
		CSROrigin: sAccount.CSROrigin,
		Cert:      sAccount.Cert,
		APIKey:    sAccount.ApiKey,
	}
}

func fromDB(dbAccount model.Account) *api.Account {
	return &api.Account{
		ID:        api.UUID(dbAccount.ID).ToBytes(),
		Name:      dbAccount.Name,
		Type:      dbAccount.Type,
		Status:    dbAccount.Status,
		CSR:       dbAccount.CSR,
		CSROrigin: dbAccount.CSROrigin,
		Cert:      dbAccount.Cert,
		ApiKey:    dbAccount.APIKey,
	}
}

// Get returns an account from its ID
func (p AccountStorage) Get(ctx context.Context, id api.UUID) (*api.Account, bool) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).GetContext(ctx, &dbAccount, uuid.UUID(id))
	if errors.Is(err, yago.ErrRecordNotFound) {
		return nil, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// GetByName returns an account with the given name if exists
func (p AccountStorage) GetByName(ctx context.Context, name string) (*api.Account, bool) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Where(
		p.model.Account.Name.Eq(name)).
		OneContext(ctx, &dbAccount)
	if errors.Is(err, yago.ErrRecordNotFound) {
		return nil, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// GetByCert returns an account with the given cert if exists
func (p AccountStorage) GetByCert(ctx context.Context, cert string) (account *api.Account, exists bool) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Where(
		p.model.Account.Cert.Eq(cert)).
		OneContext(ctx, &dbAccount)
	if errors.Is(err, yago.ErrRecordNotFound) {
		return nil, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// GetByAPIKey returns an account with the given api key if exists
func (p AccountStorage) GetByAPIKey(ctx context.Context, key string) (account *api.Account, exists bool) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	var dbAccount model.Account
	err := tx.Query(p.model.Account).Where(
		p.model.Account.APIKey.Eq(key)).
		OneContext(ctx, &dbAccount)
	if errors.Is(err, yago.ErrRecordNotFound) {
		return nil, false
	}
	if err != nil {
		panic(err)
	}
	return fromDB(dbAccount), true
}

// Create creates a new account
func (p AccountStorage) Create(ctx context.Context, account *api.Account) (*api.Account, error) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	dbAccount := toDB(account)
	if err := tx.InsertContext(ctx, &dbAccount); err != nil {
		// TODO on constraint error, return a proper error and don't panic
		panic(err)
	}
	tx.MustCommit()

	return fromDB(dbAccount), nil
}

// Delete deletes an account
func (p AccountStorage) Delete(ctx context.Context, account *api.Account) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	dbAccount := toDB(account)
	if err := tx.DeleteContext(ctx, &dbAccount); err != nil {
		panic(err)
	}
	tx.MustCommit()
}

// Update updates an account
func (p AccountStorage) Update(ctx context.Context, account *api.Account) (*api.Account, error) {
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	dbAccount := toDB(account)
	if err := tx.UpdateContext(ctx, &dbAccount); err != nil {
		// TODO on constraint error, return a proper error and don't panic
		panic(err)
	}
	tx.MustCommit()
	return fromDB(dbAccount), nil
}

// List returns a list of all the accounts
func (p AccountStorage) List(ctx context.Context) []*api.Account {
	var dbAccountList []model.Account
	tx := p.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	must(tx.Query(p.model.Account).AllContext(ctx, &dbAccountList))

	res := make([]*api.Account, len(dbAccountList))
	for i, a := range dbAccountList {
		res[i] = &api.Account{
			ID:     api.UUID(a.ID).ToBytes(),
			Name:   a.Name,
			Type:   a.Type,
			Status: a.Status,
			CSR:    a.CSR,
			Cert:   a.Cert,
		}
	}
	return res
}
