// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage/sql/model"
	"xbus.io/xbus/testutils"
	"xbus.io/xbus/testutils/sqldb"
)

// Tests around the link between actors and accounts.
func TestActorAccountLink(t *testing.T) {
	ctx := testutils.GetTestContext(t)
	// Init.
	logger := testutils.GetTestLogger(t)
	defer logger.VerifyExpectations()

	testDB := sqldb.GetTestDB(t)
	if testDB == nil {
		return
	}
	defer testDB.Close()

	m, db := testDB.GetModelDB(1)
	defer db.Close()

	actorStorage := NewActorStorage(m, db, *logger.Zero())

	// Generate a random name.
	uuidobj, _ := api.UUIDV4()
	name := uuidobj.String()

	// Create an account.
	account := model.Account{
		Name: name,
	}
	if err := db.Insert(&account); err != nil {
		t.Fatal(err)
	}

	// Create an actor.
	actor := model.Actor{
		Name: name,
		Kind: api.Actor_EMITTER,
	}
	if err := db.Insert(&actor); err != nil {
		t.Fatal(err)
	}

	// The actor -> account link should be null at this point.
	if actorInStorage, exists := actorStorage.Get(ctx, api.UUID(actor.ID)); !exists {
		t.Fatal("actor missing")
	} else {
		assert.True(t, actorInStorage.GetAccountIDAsUUID().IsZero())
	}

	// The "ListAccount" function should not find the actor so far.
	assert.Empty(t, actorStorage.ListAccount(ctx, api.UUID(account.ID)))

	// Set a link.
	assert.NoError(t, actor.AccountID.Scan(account.ID.String()))
	if err := db.Update(&actor, model.ActorAccountID); err != nil {
		t.Fatal(err)
	}

	// The actor -> account link should now exist.
	if actorInStorage, exists := actorStorage.Get(ctx, api.UUID(actor.ID)); !exists {
		t.Fatal("actor missing")
	} else {
		assert.False(t, actorInStorage.GetAccountIDAsUUID().IsZero())
	}

	// The "ListAccount" function should now find the actor.
	actorsInStorage := actorStorage.ListAccount(ctx, api.UUID(account.ID))
	assert.Len(t, actorsInStorage, 1)
	assert.Equal(t, api.UUID(account.ID), actorsInStorage[0].GetAccountIDAsUUID())

	// Ensure the account can be deleted - ONDELETE SET NULL behavior.
	if err := db.Delete(&account); err != nil {
		t.Fatal(err)
	}
	if actorInStorage, exists := actorStorage.Get(ctx, api.UUID(actor.ID)); !exists {
		t.Fatal("actor missing")
	} else {
		assert.True(t, actorInStorage.GetAccountIDAsUUID().IsZero())
	}
}
