// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"

	"github.com/orus-io/yago"
	"github.com/rs/zerolog"

	"xbus.io/xbus/storage/sql/model"
)

func must(err error) {
	if err != nil {
		panic(err)
	}
}

// TxHelper wraps a tx and make it easier to enforce commit or rollback
type TxHelper struct {
	*yago.Tx
	complete bool
	logger   zerolog.Logger
}

// MustBeginContext starts a transaction and wrap it
func MustBeginContext(ctx context.Context, db *yago.DB, logger zerolog.Logger) *TxHelper {
	tx, err := db.BeginContext(ctx)
	if err != nil {
		logger.Err(err).Msg("Cannot start a transaction")
		panic(err)
	}
	return &TxHelper{tx, false, logger}
}

// Commit commits the transaction
func (tx *TxHelper) Commit() error {
	err := tx.Tx.Commit()
	tx.complete = true
	if err != nil {
		tx.logger.Err(err).Msg("Error on commit")
	}
	return err
}

// MustCommit commits of rollback
func (tx *TxHelper) MustCommit() {
	if err := tx.Commit(); err != nil {
		panic(err)
	}
}

// Rollback rollbacks and log error if any
func (tx *TxHelper) Rollback() error {
	err := tx.Tx.Rollback()
	tx.complete = true
	if err != nil {
		tx.logger.Err(err).Msg("Error on rollback")
	}
	return err
}

// BlindRollback rollbacks and log error if any, and does not return any error
func (tx *TxHelper) BlindRollback() {
	_ = tx.Rollback()
}

// RollbackIfOpened rollback the transaction is still open
func (tx *TxHelper) RollbackIfOpened() {
	if !tx.complete {
		tx.BlindRollback()
	}
}

// Base provides a helpful base for sql-based storage implementation
type Base struct {
	model  *model.Model
	db     *yago.DB
	logger zerolog.Logger
}

// MustBeginContext stats a new transaction
func (b Base) MustBeginContext(ctx context.Context) *TxHelper {
	return MustBeginContext(ctx, b.db, b.logger)
}
