// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/qb"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/storage/sql/model"
	xuuid "xbus.io/xbus/uuid"
)

// NewLogStorage initialize a LogStorage
func NewLogStorage(model *model.Model, db *yago.DB, logger zerolog.Logger) LogStorage {
	return LogStorage{
		Base: Base{
			model: model,
			db:    db,
			logger: logger.With().
				Str("storage", "log").
				Str("implementation", "sql").
				Logger(),
		},
		tableSize: NewTableSize("log_entry"),
	}
}

// LogStorage implements storage.LogStorage
type LogStorage struct {
	Base
	tableSize *TableSize
}

// Log adds a log entry to the storage
func (s LogStorage) Log(ctx context.Context, entry *api.LogEntry) {
	tx := s.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	dbLogEntry := model.LogEntry{
		EnvelopeID: xuuid.NullUUID(entry.GetEnvelopeIDAsUUID()),
		ActorID:    xuuid.NullUUID(entry.GetActorIDAsUUID()),
		ProcessID:  xuuid.NullUUID(entry.GetProcessIDAsUUID()),
		NodeID:     entry.NodeID,
		Time:       entry.Message.Time.AsTime().UTC(),
		Level:      entry.Message.Level,
		Text:       entry.Message.Text,
	}
	if err := tx.InsertContext(ctx, &dbLogEntry); err != nil {
		panic(err)
	}
	tx.MustCommit()
}

func (s LogStorage) listContext(ctx context.Context, crits ...qb.Clause) []*api.LogEntry {
	tx := s.MustBeginContext(ctx)
	defer tx.BlindRollback()

	var dbEntries []model.LogEntry
	if err := tx.Query(s.model.LogEntry).
		Where(crits...).
		OrderBy(s.model.LogEntry.Time).
		AllContext(ctx, &dbEntries); err != nil {
		panic(err)
	}

	entries := make([]*api.LogEntry, len(dbEntries))
	for i, dbEntry := range dbEntries {
		entries[i] = &api.LogEntry{
			EnvelopeID: api.UUID(dbEntry.EnvelopeID).ToBytes(),
			ActorID:    api.UUID(dbEntry.ActorID).ToBytes(),
			ProcessID:  api.UUID(dbEntry.ProcessID).ToBytes(),
			NodeID:     dbEntry.NodeID,
			Message: &api.LogMessage{
				Time:  timestamppb.New(dbEntry.Time),
				Level: dbEntry.Level,
				Text:  dbEntry.Text,
			},
		}
	}
	return entries
}

// GetProcessMessages returns all the messages related to the given envelope
func (s LogStorage) GetProcessMessages(ctx context.Context, processID api.UUID, level api.LogLevel) []*api.LogEntry {
	crits := []qb.Clause{
		s.model.LogEntry.ProcessID.Eq(uuid.UUID(processID)),
	}

	switch level {
	case api.LogLevel_ERROR:
		crits = append(crits, s.model.LogEntry.Level.Eq(level))
	case api.LogLevel_WARNING:
		crits = append(crits, s.model.LogEntry.Level.NotEq(api.LogLevel_NOTICE))
	}

	return s.listContext(ctx, crits...)
}

// PurgeBefore removes all logs older than the given date
func (s LogStorage) PurgeBefore(ctx context.Context, t time.Time) int64 {
	tx := s.MustBeginContext(ctx)
	defer tx.MustCommit()

	t = t.UTC()

	res, err := tx.GetEngine().ExecContext(ctx,
		s.model.LogEntry.GetMapper().Table().Delete().Where(
			s.model.LogEntry.Time.Lt(t)))
	must(err)
	ra, err := res.RowsAffected()
	must(err)
	return ra
}

// Stat returns a summary of the storage content
func (s LogStorage) Stat(ctx context.Context) (*control.StorageStat_Entry, error) {
	tx := s.MustBeginContext(ctx)
	defer tx.BlindRollback()

	stat := control.StorageStat_Entry{
		Name:           "default",
		Type:           control.StorageType_LOG,
		Implementation: "pg",
	}

	if err := s.tableSize.CountSizeEngine(
		tx.GetEngine(), &stat.Count, &stat.EstimatedSize,
	); err != nil {
		return nil, err
	}

	return &stat, nil
}
