// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"
	"testing"

	"github.com/orus-io/qb"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage/sql/model"
	"xbus.io/xbus/testutils"
	"xbus.io/xbus/testutils/sqldb"
)

func TestLog(t *testing.T) {
	logger := testutils.GetTestLogger(t)
	defer logger.VerifyExpectations()

	testDB := sqldb.GetTestDB(t)
	if testDB == nil {
		return
	}
	defer testDB.Close()

	m, db := testDB.GetModelDB(1)
	defer db.Close()

	logStorage := NewLogStorage(m, db, *logger.Zero())
	db.Engine.SetLogFlags(qb.LQuery | qb.LBindings)

	t.Run("Basics", func(t *testing.T) {
		defer logger.SubTestCapture(t)()
		if _, err := db.Engine.DB().Exec("DELETE FROM log_entry"); err != nil {
			t.Fatal(err)
		}
		logentry := api.LogEntry{
			Message: &api.LogMessage{Text: "test"},
		}

		logStorage.Log(context.TODO(), &logentry)
	})

	t.Run("DeleteActor#125", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		defer logger.SubTestCapture(t)()
		if _, err := db.Engine.DB().Exec("DELETE FROM actor"); err != nil {
			t.Fatal(err)
		}
		if _, err := db.Engine.DB().Exec("DELETE FROM log_entry"); err != nil {
			t.Fatal(err)
		}

		actor := model.Actor{
			Name: "Test",
			Kind: api.Actor_EMITTER,
		}
		if err := db.Insert(&actor); err != nil {
			t.Fatal(err)
		}

		logentry := api.LogEntry{
			ActorID: api.UUID(actor.ID).ToBytes(),
			Message: &api.LogMessage{Text: "test"},
		}

		logStorage.Log(context.TODO(), &logentry)

		actorStorage := NewActorStorage(m, db, *logger.Zero())

		actorStorage.Delete(ctx, logentry.GetActorIDAsUUID())
	})

	t.Run("DeleteEnvelope", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		defer logger.SubTestCapture(t)()
		if _, err := db.Engine.DB().Exec("DELETE FROM log_entry"); err != nil {
			t.Fatal(err)
		}
		if _, err := db.Engine.DB().Exec("DELETE FROM envelope"); err != nil {
			t.Fatal(err)
		}

		envelope := model.Envelope{}
		if err := db.Insert(&envelope); err != nil {
			t.Fatal(err)
		}

		logentry := api.LogEntry{
			EnvelopeID: api.UUID(envelope.ID).ToBytes(),
			Message:    &api.LogMessage{Text: "test"},
		}

		logStorage.Log(context.TODO(), &logentry)

		envStorage := NewEnvelopeStorage(m, db, logger.Logger)

		envStorage.Purge(ctx, func(ids []api.UUID) []api.UUID { return ids })
	})
}
