// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"
	"database/sql"
	"errors"
	"fmt"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql/model"
)

// SessionStorage implements storage.SessionStorage
type SessionStorage struct {
	Base
}

// NewSessionStorage implements session storage
func NewSessionStorage(db *yago.DB, logger zerolog.Logger) *SessionStorage {
	return &SessionStorage{
		Base: Base{
			model:  nil,
			db:     db,
			logger: logger,
		},
	}
}

// List returns all the opened sessions
func (store *SessionStorage) List(ctx context.Context) ([]storage.Session, error) {
	tx := store.MustBeginContext(ctx)
	defer tx.BlindRollback()
	rows, err := tx.GetEngine().QueryContext(ctx, model.RawSQL(
		"SELECT token, account_id, valid_until FROM account_session",
	))
	if err != nil {
		return nil, err
	}
	var sessionList []storage.Session
	if err := func() (err error) {
		defer func() {
			if e := rows.Close(); e != nil && err == nil {
				err = e
			}
		}()

		for rows.Next() {
			var (
				session    = storage.Session{SessionToken: &api.SessionToken{}}
				validUntil time.Time
			)
			if err := rows.Scan(
				&session.Token,
				(*uuid.UUID)(&session.AccountID),
				&validUntil,
			); err != nil {
				return err
			}
			session.ValidUntil = timestamppb.New(validUntil.UTC())
			sessionList = append(sessionList, session)
		}
		return nil
	}(); err != nil {
		return nil, err
	}
	return sessionList, nil
}

// Get returns a Session from its token
func (store *SessionStorage) Get(ctx context.Context, token string) (*storage.Session, error) {
	tx := store.MustBeginContext(ctx)
	defer tx.BlindRollback()
	row := tx.GetEngine().QueryRowContext(ctx, model.RawSQL(
		"SELECT token, account_id, valid_until "+
			"FROM account_session "+
			"WHERE token = $1",
		token,
	))
	var (
		session    = storage.Session{SessionToken: &api.SessionToken{}}
		validUntil time.Time
	)
	if err := row.Scan(
		&session.Token,
		(*uuid.UUID)(&session.AccountID),
		&validUntil,
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, fmt.Errorf("%w: session '%s'", storage.ErrNotExist, token)
		}
		return nil, err
	}
	session.ValidUntil = timestamppb.New(validUntil.UTC())
	return &session, nil
}

// Set create or update a Session
func (store *SessionStorage) Set(ctx context.Context, session storage.Session) error {
	tx := store.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	row := tx.GetEngine().QueryRowContext(ctx, model.RawSQL(
		"INSERT INTO account_session (token, account_id, valid_until) "+
			"VALUES ($1, $2, $3) "+
			"ON CONFLICT (token) "+
			"DO UPDATE SET valid_until = $3 "+
			"RETURNING (account_id)",
		session.Token,
		uuid.UUID(session.AccountID),
		session.ValidUntil.AsTime(),
	))

	var accountID uuid.UUID

	if err := row.Scan(&accountID); err != nil {
		return err
	}

	if accountID != uuid.UUID(session.AccountID) {
		return fmt.Errorf(
			"Token %s already exists and is not associated to the account %s, "+
				"but to the account %s",
			session.Token, session.AccountID, accountID,
		)
	}

	return tx.Commit()
}

// Delete deletes a token
func (store *SessionStorage) Delete(ctx context.Context, token string) error {
	tx := store.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().ExecContext(ctx, model.RawSQL(
		"DELETE FROM account_session WHERE token = $1",
		token,
	))
	if err != nil {
		return err
	}

	return tx.Commit()
}

// DeleteOlderThan deletes the sessions based on their 'validUntil'
func (store *SessionStorage) DeleteOlderThan(ctx context.Context, time time.Time) error {
	tx := store.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().ExecContext(ctx, model.RawSQL(
		"DELETE FROM account_session "+
			"WHERE valid_until < $1",
		time,
	))
	if err != nil {
		return err
	}

	return tx.Commit()
}
