// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"database/sql"

	"github.com/orus-io/yago"

	"xbus.io/xbus/storage/sql/model"
)

type queryRow interface {
	QueryRow(query string, args ...interface{}) *sql.Row
}

type TableSize struct {
	tablename string
	fast      bool
}

func NewTableSize(tablename string) *TableSize {
	return &TableSize{
		tablename: tablename,
		fast:      true,
	}
}

func (s *TableSize) Count(queryRow queryRow, count *uint64) error {
	if s.fast {
		if err := queryRow.QueryRow(
			"SELECT n_live_tup FROM pg_stat_all_tables WHERE relname = $1",
			s.tablename,
		).Scan(count); err != nil {
			return err
		}
		if *count < 1000*1000 {
			s.fast = false
		}
	}

	if !s.fast {
		if err := queryRow.QueryRow(
			"SELECT COUNT(*) FROM " + s.tablename,
		).Scan(count); err != nil {
			return err
		}
		if *count < 1000*1000 {
			s.fast = false
		}
	}

	return nil
}

func (s *TableSize) Size(queryRow queryRow, size *uint64) error {
	if err := queryRow.
		QueryRow("SELECT pg_total_relation_size($1)", s.tablename).
		Scan(size); err != nil {
		return err
	}
	return nil
}

func (s *TableSize) CountSize(queryRow queryRow, count, size *uint64) error {
	if err := s.Count(queryRow, count); err != nil {
		return err
	}
	if err := s.Size(queryRow, size); err != nil {
		return err
	}
	return nil
}

func (s *TableSize) CountEngine(engine yago.Engine, count *uint64) error {
	if s.fast {
		if err := engine.QueryRow(
			model.RawSQL(
				"SELECT n_live_tup FROM pg_stat_all_tables WHERE relname = $1",
				s.tablename,
			)).
			Scan(count); err != nil {
			return err
		}
		if *count < 1000*1000 {
			s.fast = false
		}
	}

	if !s.fast {
		if err := engine.
			QueryRow(
				model.RawSQL("SELECT COUNT(*) FROM " + s.tablename),
			).
			Scan(count); err != nil {
			return err
		}
		if *count > 1200*1000 {
			s.fast = true
		}
	}

	return nil
}

func (s *TableSize) SizeEngine(engine yago.Engine, size *uint64) error {
	if err := engine.
		QueryRow(model.RawSQL("SELECT pg_total_relation_size($1)", s.tablename)).
		Scan(size); err != nil {
		return err
	}
	return nil
}

func (s *TableSize) CountSizeEngine(engine yago.Engine, count, size *uint64) error {
	if err := s.CountEngine(engine, count); err != nil {
		return err
	}
	if err := s.SizeEngine(engine, size); err != nil {
		return err
	}
	return nil
}
