// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package withfallback

import (
	"context"

	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/storage"
)

func NewEnvelopeStorage(
	primary, fallback storage.EnvelopeStorage,
) EnvelopeStorage {
	return EnvelopeStorage{
		primary, fallback,
	}
}

type EnvelopeStorage struct {
	storage.EnvelopeStorage
	fallback storage.EnvelopeStorage
}

func (s EnvelopeStorage) GetEventTypes(ctx context.Context, id api.UUID) ([]string, bool) {
	types, exists := s.EnvelopeStorage.GetEventTypes(ctx, id)
	if !exists {
		types, exists = s.fallback.GetEventTypes(ctx, id)
	}
	return types, exists
}

func (s EnvelopeStorage) ReceptionStatus(ctx context.Context, id api.UUID) (storage.EnvelopeStorageReceptionStatus, error) {
	status, err := s.EnvelopeStorage.ReceptionStatus(ctx, id)
	// TODO we need a ErrNoEnvelope error to avoid skipping important errors
	if err != nil {
		status, err = s.fallback.ReceptionStatus(ctx, id)
	}
	return status, err
}

func (s EnvelopeStorage) ReadEnvelope(
	ctx context.Context, id api.UUID, position storage.EnvelopeReadPos, maxsize int,
) (*api.Envelope, storage.EnvelopeReadPos, error) {
	e, p, err := s.EnvelopeStorage.ReadEnvelope(ctx, id, position, maxsize)
	if err != nil {
		e, p, err = s.fallback.ReadEnvelope(ctx, id, position, maxsize)
	}
	return e, p, err
}

func (s EnvelopeStorage) Purge(ctx context.Context, filter storage.EnvelopeFilter) {
	s.EnvelopeStorage.Purge(ctx, filter)
	s.fallback.Purge(ctx, filter)
}

func (s EnvelopeStorage) Stat(ctx context.Context) ([]*control.StorageStat_Entry, error) {
	primary, err := s.EnvelopeStorage.Stat(ctx)
	if err != nil {
		return nil, err
	}
	fallback, err := s.fallback.Stat(ctx)
	if err != nil {
		return nil, err
	}

	stats := make([]*control.StorageStat_Entry, 0, len(primary))
	stats = append(stats, primary...)
	stats = append(stats, fallback...)

	total := control.StorageStat_Entry{
		Implementation: "withfallback",
	}
	total.Name = "("
	for i, s := range primary {
		if i != 0 {
			total.Name += ", "
		}
		total.Name += s.Name
	}
	total.Name += ") withfallback ("
	for i, s := range fallback {
		if i != 0 {
			total.Name += ", "
		}
		total.Name += s.Name
	}
	total.Name += ")"

	stats = append(stats, &total)
	return stats, nil
}
