// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package withfallback

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/testutils"
	"xbus.io/xbus/testutils/mocks"
	storagetest "xbus.io/xbus/testutils/storage"
)

func envelopeStorageFactory(tb testing.TB) (storage.EnvelopeStorage, func()) {
	tb.Helper()

	return NewEnvelopeStorage(
		mocks.NewEnvelopeStorage(testutils.GetTestLogger(tb).Logger),
		mocks.NewEnvelopeStorage(testutils.GetTestLogger(tb).Logger),
	), func() {}
}

func TestEnvelopeStorage(t *testing.T) {
	storagetest.EnvelopeStorageTests(t, envelopeStorageFactory, storagetest.EnvelopeStorageTestsOptions{
		SupportUnordered:     true,
		SupportParallelStore: true,
	})

	t.Run("fallback", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		primary := mocks.NewEnvelopeStorage(testutils.GetTestLogger(t).Logger)
		fallback := mocks.NewEnvelopeStorage(testutils.GetTestLogger(t).Logger)

		combined := NewEnvelopeStorage(primary, fallback)

		env1 := &api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			EventIDs: api.UUIDSlice{
				api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
			}.ToBytes(),
		}

		env2 := &api.Envelope{
			ID: env1.ID,
			Events: []*api.Event{
				{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index: 1,
					Type:  "xbus.test",
					Items: [][]byte{
						[]byte("Data"),
						[]byte("Data"),
					},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		_, err := fallback.StoreEnvelope(ctx, env1)
		assert.NoError(t, err)

		status, err := combined.ReceptionStatus(ctx, env1.GetIDAsUUID())
		assert.NoError(t, err)
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status)

		_, err = fallback.StoreEnvelope(ctx, env2)
		assert.NoError(t, err)

		status, err = combined.ReceptionStatus(ctx, env1.GetIDAsUUID())
		assert.NoError(t, err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status)

		_, _, err = primary.ReadEnvelope(ctx, env1.GetIDAsUUID(), storage.EnvelopeStartPos(), 0)
		assert.ErrorIs(t, err, storage.ErrNoSuchID)

		_, _, err = combined.ReadEnvelope(ctx, env1.GetIDAsUUID(), storage.EnvelopeStartPos(), 0)
		assert.NoError(t, err)
	})
}
