// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package storage

import (
	"context"
	"errors"
	"sync"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/testutils"
)

// EnvelopeStorageFactory is any factory of EnvelopeStorage
type EnvelopeStorageFactory func(bt testing.TB) (storage.EnvelopeStorage, func())

type EnvelopeStorageTestsOptions struct {
	SupportUnordered     bool
	SupportParallelStore bool
}

// EnvelopeStorageTests is a test suite for EnvelopeStorage implementations
func EnvelopeStorageTests(t *testing.T, factory EnvelopeStorageFactory, options EnvelopeStorageTestsOptions) {
	t.Helper()
	// setup a little event handler mock
	var (
		handlerCalled       bool
		handlerParamID      api.UUID
		handlerParamStatus  storage.EnvelopeStorageReceptionStatus
		handlerParamNewData bool
	)

	handler := func(id api.UUID, status storage.EnvelopeStorageReceptionStatus, newData bool) {
		handlerCalled = true
		handlerParamID = id
		handlerParamStatus = status
		handlerParamNewData = newData
	}

	resetHandler := func() {
		handlerCalled = false
	}

	t.Run("Store/NoSuchID", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()

		id := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")

		_, err := envelopeStorage.ReceptionStatus(ctx, id)
		assert.ErrorIs(t, err, storage.ErrNoSuchID)

		_, _, err = envelopeStorage.ReadEnvelope(ctx, id, storage.EnvelopeStartPos(), 0)
		assert.ErrorIs(t, err, storage.ErrNoSuchID)
	})

	t.Run("Store/TypesUnknown", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()

		env := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		status, err := envelopeStorage.StoreEnvelope(ctx, &env)
		assert.NoError(t, err)
		assert.Equal(t, env.GetIDAsUUID(), status.ID)
		assert.False(t, status.TypesKnown)
		assert.Equal(t, 1, len(status.Events))
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())
		if t.Failed() {
			return
		}
	})

	t.Run("NullFirstFragmentThenComplete", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		if envelopeStorage == nil {
			t.Fatal("on envelope storage")
		}
		defer clean()

		env1 := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
		}

		env2 := api.Envelope{
			ID: env1.ID,
			EventIDs: [][]byte{
				api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
			},
			Events: []*api.Event{
				{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index: 1,
					Type:  "xbus.test",
					Items: [][]byte{
						[]byte("Data"),
						[]byte("Data"),
					},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		status, err := envelopeStorage.StoreEnvelope(ctx, &env1)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())

		rStatus, err := envelopeStorage.ReceptionStatus(ctx, env1.GetIDAsUUID())
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, rStatus)

		status, err = envelopeStorage.StoreEnvelope(ctx, &env2)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Status.String(), "%#v", status)

		rStatus, err = envelopeStorage.ReceptionStatus(ctx, env1.GetIDAsUUID())
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, rStatus)
	})

	t.Run("DataLessFirstFragment", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		if envelopeStorage == nil {
			t.Fatal("on envelope storage")
		}
		defer clean()

		env1 := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			EventIDs: api.UUIDSlice{
				api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
			}.ToBytes(),
		}

		env2 := api.Envelope{
			ID: env1.ID,
			Events: []*api.Event{
				{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index: 1,
					Type:  "xbus.test",
					Items: [][]byte{
						[]byte("Data"),
						[]byte("Data"),
					},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		status, err := envelopeStorage.StoreEnvelope(ctx, &env1)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())

		rStatus, err := envelopeStorage.ReceptionStatus(ctx, env1.GetIDAsUUID())
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, rStatus)

		status, err = envelopeStorage.StoreEnvelope(ctx, &env2)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Status.String(), "%#v", status)

		rStatus, err = envelopeStorage.ReceptionStatus(ctx, env1.GetIDAsUUID())
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, rStatus)
	})

	t.Run("Store/SingleEnvelope", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		if envelopeStorage == nil {
			t.Fatal("failed to init storage")
		}
		defer clean()
		envelopeStorage.SetEventHandler(handler)

		env := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].GetIDAsUUID()
		env.EventIDs = api.UUIDSlice{eventID}.ToBytes()

		resetHandler()
		status, err := envelopeStorage.StoreEnvelope(ctx, &env)
		assert.Nil(t, err)
		assert.Equal(t, env.GetIDAsUUID(), status.ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, uint64(1), status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

		assert.True(t, handlerCalled)
		assert.Equal(t, env.GetIDAsUUID(), handlerParamID)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, handlerParamStatus)
		assert.True(t, handlerParamNewData)

		types, exists := envelopeStorage.GetEventTypes(ctx, env.GetIDAsUUID())
		assert.True(t, exists)
		assert.Equal(t, []string{"xbus.test"}, types)
	})

	t.Run("Store/PairOfEnvelope", func(t *testing.T) {
		env1 := api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index: 1,
					Type:  "xbus.test",
					Items: [][]byte{[]byte("Data")},
				},
			},
		}

		eventID := env1.Events[0].GetIDAsUUID()
		env1.EventIDs = api.UUIDSlice{eventID}.ToBytes()

		env2 := api.Envelope{
			ID: env1.ID,
			Events: []*api.Event{
				{
					ID:        eventID.ToBytes(),
					Index:     2,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		t.Run("Ordered", func(t *testing.T) {
			ctx := testutils.GetTestContext(t)
			envelopeStorage, clean := factory(t)
			defer clean()
			envelopeStorage.SetEventHandler(handler)

			resetHandler()
			status, err := envelopeStorage.StoreEnvelope(ctx, &env1)
			assert.Nil(t, err)
			assert.Equal(t, env1.GetIDAsUUID(), status.ID)
			if len(status.Events) != 1 {
				t.Fatalf("Expected 1 event, got %d", len(status.Events))
			}
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.True(t, status.TypesKnown)
			assert.Equal(t, uint64(1), status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Events[0].Status)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())

			assert.True(t, handlerCalled)
			assert.Equal(t, env1.GetIDAsUUID(), handlerParamID)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, handlerParamStatus)
			assert.True(t, handlerParamNewData)

			resetHandler()
			status, err = envelopeStorage.StoreEnvelope(ctx, &env2)
			assert.NoError(t, err)
			assert.Equal(t, env2.GetIDAsUUID(), status.ID)
			assert.Equal(t, 1, len(status.Events))
			if t.Failed() {
				return
			}
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.Equal(t, uint64(2), status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

			assert.True(t, handlerCalled)
			assert.Equal(t, env1.GetIDAsUUID(), handlerParamID)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, handlerParamStatus)
			assert.True(t, handlerParamNewData)
		})

		t.Run("Unordered", func(t *testing.T) {
			ctx := testutils.GetTestContext(t)
			if !options.SupportUnordered {
				t.Skip("not supported by this implementation")
			}
			envelopeStorage, clean := factory(t)
			defer clean()

			status, err := envelopeStorage.StoreEnvelope(ctx, &env2)
			assert.Nil(t, err)
			assert.Equal(t, env1.GetIDAsUUID(), status.ID)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(ctx, &env1)
			assert.Nil(t, err)
			assert.Equal(t, env2.GetIDAsUUID(), status.ID)
			assert.Equal(t, 1, len(status.Events))
			if t.Failed() {
				return
			}
			assert.Equal(t, eventID, status.Events[0].ID)
			assert.True(t, status.TypesKnown)
			assert.Equal(t, uint64(2), status.Events[0].AvailableItems)
			assert.Equal(t, nil, status.Events[0].Err)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Events[0].Status.String())
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Status.String())
		})

		t.Run("CheckStalled", func(t *testing.T) {
			ctx := testutils.GetTestContext(t)
			envelopeStorage, clean := factory(t)
			defer clean()
			envelopeStorage.SetEventHandler(handler)

			status, err := envelopeStorage.StoreEnvelope(ctx, &env1)
			if !assert.NoError(t, err) {
				t.FailNow()
			}
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			assert.False(t, envelopeStorage.CheckStalled(ctx, env1.GetIDAsUUID(), 200*time.Millisecond))

			time.Sleep(100 * time.Millisecond)
			resetHandler()
			assert.True(t, envelopeStorage.CheckStalled(ctx, env1.GetIDAsUUID(), time.Millisecond))

			assert.True(t, handlerCalled)
			assert.Equal(t, env1.GetIDAsUUID(), handlerParamID)
			assert.Equal(t, storage.EnvelopeStorageReceptionStalled, handlerParamStatus)
			assert.False(t, handlerParamNewData)

			resetHandler()
			assert.True(t, envelopeStorage.CheckStalled(ctx, env1.GetIDAsUUID(), time.Second))
			assert.False(t, handlerCalled)

			resetHandler()
			status, err = envelopeStorage.StoreEnvelope(ctx, &env2)
			assert.Nil(t, err)
			assert.Equal(t, env2.GetIDAsUUID(), status.ID)
			assert.False(t, envelopeStorage.CheckStalled(ctx, env1.GetIDAsUUID(), time.Second))
		})
	})

	t.Run("Read/SingleEnvelope", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()
		env := &api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].GetIDAsUUID()
		env.EventIDs = api.UUIDSlice{eventID}.ToBytes()

		status, err := envelopeStorage.StoreEnvelope(ctx, env)
		assert.Nil(t, err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(ctx, env.GetIDAsUUID(), storage.EnvelopeStartPos(), 0)
		assert.Nil(t, err)
		assert.True(t, pos.Complete)
		assert.Equal(t, env.ID, out.ID)
		assert.Equal(t, 1, len(out.Events))
		if t.Failed() {
			return
		}
		outEvent := out.Events[0]
		assert.Equal(t, eventID, outEvent.GetIDAsUUID())
		assert.Equal(t, uint64(1), outEvent.Index)
		assert.Equal(t, "xbus.test", outEvent.Type)
		assert.Equal(t, [][]byte{[]byte("Data")}, outEvent.Items)
		assert.Equal(t, uint64(1), outEvent.ItemCount)
		assert.Equal(t, api.Checksum(1859455373), api.Checksum(outEvent.Checksum))
	})

	t.Run("Stream/PairOfEnvelope", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()

		env := &api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:    api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index: 1,
					Type:  "xbus.test",
					Items: [][]byte{[]byte("Data")},
				},
			},
		}

		eventID := env.Events[0].GetIDAsUUID()
		env.EventIDs = api.UUIDSlice{eventID}.ToBytes()

		status, err := envelopeStorage.StoreEnvelope(ctx, env)
		assert.Nil(t, err)
		assert.Equal(t, env.GetIDAsUUID(), status.ID)
		if len(status.Events) != 1 {
			t.Fatalf("Expected 1 event, got %d", len(status.Events))
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, uint64(0), status.Events[0].TotalItems)
		assert.Equal(t, uint64(1), status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Events[0].Status)
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(
			ctx, env.GetIDAsUUID(), storage.EnvelopeStartPos(), 0)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, 1, len(out.Events))
		assert.Equal(t, 1, len(out.Events[0].Items))
		assert.Equal(t, "Data", string(out.Events[0].Items[0]))

		env = &api.Envelope{
			ID: env.ID,
			Events: []*api.Event{
				{
					ID:        eventID.ToBytes(),
					Index:     2,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data2")},
					ItemCount: 2,
					Checksum:  0x92b3a8e2,
				},
			},
		}

		env.EventIDs = api.UUIDSlice{eventID}.ToBytes()

		status, err = envelopeStorage.StoreEnvelope(ctx, env)
		assert.Nil(t, err)
		assert.Equal(t, env.GetIDAsUUID(), status.ID)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, uint64(2), status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Events[0].Status)
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)

		out, _, err = envelopeStorage.ReadEnvelope(ctx, env.GetIDAsUUID(), pos, 0)
		assert.NoError(t, err)
		if assert.Equal(t, 1, len(out.Events)) {
			assert.Equal(t, "Data2", string(out.Events[0].Items[0]))
		}
	})

	t.Run("Stream/ItemLessFragments", func(t *testing.T) {
		envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
		eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")

		fragments := []*api.Envelope{
			{
				ID:       envID.ToBytes(),
				EventIDs: api.UUIDSlice{eventID}.ToBytes(),
				Events: []*api.Event{
					{
						ID:    eventID.ToBytes(),
						Index: 1,
						Type:  "xbus.test",
					},
				},
			},
			{
				ID: envID.ToBytes(),
				Events: []*api.Event{
					{
						ID:    eventID.ToBytes(),
						Index: 2,
						Type:  "xbus.test",
						Items: [][]byte{[]byte("item1")},
					},
				},
			},
			{
				ID: envID.ToBytes(),
				Events: []*api.Event{
					{
						ID:    eventID.ToBytes(),
						Index: 3,
						Type:  "xbus.test",
					},
				},
			},
			{
				ID: envID.ToBytes(),
				Events: []*api.Event{
					{
						ID:        eventID.ToBytes(),
						Index:     4,
						ItemCount: 2,
						Type:      "xbus.test",
						Items:     [][]byte{[]byte("item2")},
						Checksum:  0x220c46c7,
					},
				},
			},
		}

		bigEnvFragments := GenBigEnvFragments()

		t.Run("Ordered", func(t *testing.T) {
			ctx := testutils.GetTestContext(t)
			envelopeStorage, clean := factory(t)
			defer clean()

			status, err := envelopeStorage.StoreEnvelope(ctx, fragments[0])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(ctx, fragments[1])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(ctx, fragments[2])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

			status, err = envelopeStorage.StoreEnvelope(ctx, fragments[3])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)
			assert.Equal(t, "", status.Error())

			out := readComplete(ctx, t, envelopeStorage, envID)
			assert.Equal(t, 1, len(out.Events))
			if len(out.Events) == 1 {
				assert.Equal(t, [][]byte{api.Item("item1"), api.Item("item2")}, out.Events[0].Items)
			}
		})

		t.Run("Ordered/ReadBig", func(t *testing.T) {
			ctx := testutils.GetTestContext(t)
			envelopeStorage, clean := factory(t)
			defer clean()

			for _, fragment := range bigEnvFragments[:4] {
				status, err := envelopeStorage.StoreEnvelope(ctx, fragment)
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)
			}

			var wg sync.WaitGroup
			wg.Add(1)
			defer wg.Wait()

			go func() {
				defer wg.Done()

				time.Sleep(100 * time.Millisecond)
				for _, fragment := range bigEnvFragments[4 : len(bigEnvFragments)-1] {
					status, err := envelopeStorage.StoreEnvelope(ctx, fragment)
					assert.Nil(t, err)
					assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)
				}
				status, err := envelopeStorage.StoreEnvelope(ctx, bigEnvFragments[len(bigEnvFragments)-1])
				assert.Nil(t, err)
				assert.Equal(t, storage.EnvelopeStorageReceptionComplete, status.Status)
				assert.Equal(t, "", status.Error())
			}()

			pos := storage.EnvelopeStartPos()
			itemCount := 0
			fragmentCount := 0
			consecutiveWaits := 0
			for !pos.Complete {
				out, newPos, err := envelopeStorage.ReadEnvelope(ctx, envID, pos, 500)
				if errors.Is(err, storage.ErrNoSuchID) {
					time.Sleep(1 * time.Millisecond)
					continue
				}
				if err != nil {
					t.Fatal(err)
				}
				if newPos.Equals(pos) {
					consecutiveWaits++
					if consecutiveWaits >= 10 {
						t.Fatal("Waited too long for the next fragment", fragmentCount)
					}
					time.Sleep(100 * time.Millisecond)
				} else {
					consecutiveWaits = 0
					fragmentCount++
				}

				for _, ev := range out.Events {
					itemCount += len(ev.Items)
					if newPos.Events[eventID].ItemCount > 202 {
						t.Fatal("ReadEnvelope returning too many items")
					}
				}
				pos = newPos
			}
			assert.Equal(t, 202, itemCount)
		})

		t.Run("Unordered", func(t *testing.T) {
			ctx := testutils.GetTestContext(t)
			if !options.SupportUnordered {
				t.Skip("not supported by this implementation")
			}
			envelopeStorage, clean := factory(t)
			defer clean()

			status, err := envelopeStorage.StoreEnvelope(ctx, fragments[2])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())

			status, err = envelopeStorage.StoreEnvelope(ctx, fragments[0])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())

			status, err = envelopeStorage.StoreEnvelope(ctx, fragments[3])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionReceiving.String(), status.Status.String())

			status, err = envelopeStorage.StoreEnvelope(ctx, fragments[1])
			assert.Nil(t, err)
			assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Status.String())

			out := readComplete(ctx, t, envelopeStorage, envID)
			assert.Equal(t, 1, len(out.Events))
			if len(out.Events) == 1 {
				assert.Equal(t, [][]byte{api.Item("item1"), api.Item("item2")}, out.Events[0].Items)
			}
		})
	})

	t.Run("Stream/EarlyEventCompletion", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()

		envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
		eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")
		eventID2 := api.MustUUIDFromString("615d146a-02af-11e7-8cdf-d3cb88225aff")

		status, err := envelopeStorage.StoreEnvelope(
			ctx,
			&api.Envelope{
				ID:       envID.ToBytes(),
				EventIDs: api.UUIDSlice{eventID, eventID2}.ToBytes(),
				Events: []*api.Event{
					{
						ID:        eventID.ToBytes(),
						Type:      "xbus.test",
						Index:     1,
						Items:     [][]byte{[]byte("Data")},
						ItemCount: 1,
						Checksum:  0x6ed5098d,
					},
				},
			})
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionReceiving, status.Status)

		status, err = envelopeStorage.StoreEnvelope(
			ctx,
			&api.Envelope{
				ID:       envID.ToBytes(),
				EventIDs: api.UUIDSlice{eventID, eventID2}.ToBytes(),
				Events: []*api.Event{
					{
						ID:        eventID2.ToBytes(),
						Type:      "xbus.test.2",
						Index:     1,
						Items:     [][]byte{[]byte("Data")},
						ItemCount: 1,
						Checksum:  0x15d47408,
					},
				},
				Last: true,
			})
		if !assert.NoError(t, err) {
			t.FailNow()
		}
		assert.Equal(t, storage.EnvelopeStorageReceptionComplete.String(), status.Status.String())
	})

	t.Run("Purge", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()
		env1 := &api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		env2 := &api.Envelope{
			ID: api.MustUUIDFromString("7f107142-8977-11e7-bd71-635f1840d117").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		env1.EventIDs = api.UUIDSlice{env1.Events[0].GetIDAsUUID()}.ToBytes()
		env2.EventIDs = api.UUIDSlice{env2.Events[0].GetIDAsUUID()}.ToBytes()

		_, err := envelopeStorage.StoreEnvelope(ctx, env1)
		assert.Nil(t, err)

		_, err = envelopeStorage.StoreEnvelope(ctx, env2)
		assert.Nil(t, err)

		_, exists := envelopeStorage.GetEventTypes(ctx, env1.GetIDAsUUID())
		assert.True(t, exists)

		_, exists = envelopeStorage.GetEventTypes(ctx, env2.GetIDAsUUID())
		assert.True(t, exists)

		envelopeStorage.Purge(ctx, func(ids []api.UUID) []api.UUID {
			var purge api.UUIDSlice
			for _, id := range ids {
				if id == env1.GetIDAsUUID() {
					purge = append(purge, id)
				}
			}
			return purge
		})

		_, exists = envelopeStorage.GetEventTypes(ctx, env1.GetIDAsUUID())
		assert.False(t, exists)
		_, exists = envelopeStorage.GetEventTypes(ctx, env2.GetIDAsUUID())
		assert.True(t, exists)
	})

	t.Run("Parallel_store", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		if !options.SupportParallelStore {
			t.Skip("not supported by this implementation")
		}
		envelopeStorage, clean := factory(t)
		defer clean()

		item := []byte("someitem,\n")

		var wg sync.WaitGroup
		for i := 0; i != 10; i++ {
			wg.Add(1)
			go func() {
				defer wg.Done()

				for i := 0; i != 2; i++ {
					sender := xbus.NewEnvelopeSender(
						[]string{"xbus.test"},
						1024*100,
						func(env *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
							state, err := envelopeStorage.StoreEnvelope(ctx, env)

							if !assert.NoError(t, err) {
								return api.EnvelopeAck_ERROR, err
							}

							if state.Status == storage.EnvelopeStorageReceptionComplete {
								return api.EnvelopeAck_ACCEPTED, nil
							}
							return api.EnvelopeAck_RECEIVING, nil
						})

					for i := 0; i != 100000; i++ {
						if !assert.NoError(t, sender.AddItems("xbus.test", item)) {
							return
						}
					}
					if !assert.NoError(t, sender.Close()) {
						return
					}
				}
			}()
		}
		wg.Wait()
	})

	t.Run("Summary", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		envelopeStorage, clean := factory(t)
		defer clean()
		env1 := &api.Envelope{
			ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		env2 := &api.Envelope{
			ID: api.MustUUIDFromString("7f107142-8977-11e7-bd71-635f1840d117").ToBytes(),
			Events: []*api.Event{
				{
					ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
					Index:     1,
					Type:      "xbus.test",
					Items:     [][]byte{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		_, err := envelopeStorage.StoreEnvelope(ctx, env1)
		require.Nil(t, err)

		_, err = envelopeStorage.StoreEnvelope(ctx, env2)
		require.Nil(t, err)

		stat, err := envelopeStorage.Stat(ctx)
		require.Nil(t, err)

		assert.EqualValues(t, 2, stat[0].Count)

		// The estimateSize can be a lot bigger that the actual envelope sizes,
		// depending on how the storage works.
		assert.GreaterOrEqual(t, stat[0].EstimatedSize, uint64(130))
	})
}

// GenBigEnvFragments creates fragments of a big envelope
func GenBigEnvFragments() []*api.Envelope {
	envID := api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df")
	eventID := api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882")
	eventID2 := api.MustUUIDFromString("615d146a-02af-11e7-8cdf-d3cb88225aff")
	bigEnvFragments := []*api.Envelope{
		{
			ID:       envID.ToBytes(),
			EventIDs: api.UUIDSlice{eventID, eventID2}.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID.ToBytes(),
					Index: 1,
					Type:  "xbus.test",
				},
				{
					ID:    eventID2.ToBytes(),
					Index: 1,
					Type:  "xbus.test.2",
				},
			},
		},
		{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID.ToBytes(),
					Index: 2,
					Type:  "xbus.test",
				},
			},
		},
		{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID.ToBytes(),
					Index: 3,
					Type:  "xbus.test",
				},
			},
		},
		{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID.ToBytes(),
					Index: 4,
					Type:  "xbus.test",
				},
			},
		},
	}

	item := make([]byte, 100)

	for i := 0; i != 100; i++ {
		bigEnvFragments = append(bigEnvFragments, &api.Envelope{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID.ToBytes(),
					Index: uint64(i + 5),
					Type:  "xbus.test",
					Items: [][]byte{item},
				},
			},
		})
	}
	for i := 0; i != 100; i++ {
		bigEnvFragments = append(bigEnvFragments, &api.Envelope{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID2.ToBytes(),
					Index: uint64(i + 2),
					Type:  "xbus.test.2",
					Items: [][]byte{item},
				},
			},
		})
	}

	bigEnvFragments = append(bigEnvFragments,
		&api.Envelope{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:    eventID.ToBytes(),
					Index: 105,
					Type:  "xbus.test",
				},
			},
		},
		&api.Envelope{
			ID: envID.ToBytes(),
			Events: []*api.Event{
				{
					ID:        eventID.ToBytes(),
					Index:     106,
					Type:      "xbus.test",
					Items:     [][]byte{item},
					ItemCount: 101,
					Checksum:  0xbdac9720,
				},
				{
					ID:        eventID2.ToBytes(),
					Index:     102,
					Type:      "xbus.test.2",
					Items:     [][]byte{item},
					ItemCount: 101,
					Checksum:  0x560279ee,
				},
			},
		})
	return bigEnvFragments
}

func readComplete(ctx context.Context, t *testing.T, envelopeStorage storage.EnvelopeStorage, envID api.UUID) *api.Envelope {
	t.Helper()
	out, pos, err := envelopeStorage.ReadEnvelope(ctx, envID, storage.EnvelopeStartPos(), 0)
	require.NoError(t, err)
	for !pos.Complete {
		fragment, npos, err := envelopeStorage.ReadEnvelope(ctx, envID, pos, 0)
		require.NoError(t, err)
		require.NoError(t, out.Merge(fragment))
		pos = npos
	}
	require.True(t, pos.Complete)
	return out
}
