// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package storage

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/testutils"
)

var simpleEnvelope = api.Envelope{
	ID: api.MustUUIDFromString("b74415ee-83c1-11e6-8f04-133b336287df").ToBytes(),
	Events: []*api.Event{
		{
			ID:        api.MustUUIDFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882").ToBytes(),
			Index:     1,
			Type:      "xbus.test",
			Items:     [][]byte{[]byte("Data")},
			ItemCount: 1,
			Checksum:  1859455373,
		},
	},
}

// EnvelopeStorageBench is a benchmark suite for a EnvelopeStorage implementation
func EnvelopeStorageBench(b *testing.B, factory EnvelopeStorageFactory) {
	b.Helper()

	b.Run("InitClearStorage", func(b *testing.B) {
		for i := 0; i != b.N; i++ {
			_, clean := factory(b)
			clean()
		}
	})
	b.Run("StoreAndReadOneSimpleEnvelope", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		b.StopTimer()
		for i := 0; i != b.N; i++ {
			es, clean := factory(b)
			b.StartTimer()
			_, err := es.StoreEnvelope(ctx, &simpleEnvelope)
			if !assert.NoError(b, err) {
				b.FailNow()
			}
			_, _, err = es.ReadEnvelope(ctx, simpleEnvelope.GetIDAsUUID(), storage.EnvelopeStartPos(), 1024*1024)
			if !assert.NoError(b, err) {
				b.FailNow()
			}
			b.StopTimer()
			clean()
		}
	})
	b.Run("ReadSimpleEnvelope", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		es, clean := factory(b)
		defer clean()

		_, err := es.StoreEnvelope(ctx, &simpleEnvelope)
		require.NoError(b, err)

		b.ResetTimer()

		for i := 0; i != b.N; i++ {
			_, _, err := es.ReadEnvelope(ctx, simpleEnvelope.GetIDAsUUID(), storage.EnvelopeStartPos(), 1024*1024)
			require.NoError(b, err)
		}
	})
	b.Run("StoreOnceReadFiveSimpleEnvelope", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		b.StopTimer()
		for i := 0; i != b.N; i++ {
			es, clean := factory(b)
			b.StartTimer()
			_, err := es.StoreEnvelope(ctx, &simpleEnvelope)
			if !assert.NoError(b, err) {
				b.FailNow()
			}
			for i := 0; i <= 5; i++ {
				_, _, err := es.ReadEnvelope(ctx, simpleEnvelope.GetIDAsUUID(), storage.EnvelopeStartPos(), 1024*1024)
				if !assert.NoError(b, err) {
					b.FailNow()
				}
			}
			b.StopTimer()
			clean()
		}
	})
	b.Run("StoreAndReadManySimpleEnvelope", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		es, clean := factory(b)
		defer clean()
		for i := 0; i != b.N; i++ {
			id, err := api.UUIDV4()
			require.NoError(b, err)
			simpleEnvelope.ID = id.ToBytes()
			if err != nil {
				panic(err)
			}
			_, err = es.StoreEnvelope(ctx, &simpleEnvelope)
			require.NoError(b, err)
			_, _, err = es.ReadEnvelope(ctx, simpleEnvelope.GetIDAsUUID(), storage.EnvelopeStartPos(), 1024*1024)
			require.NoError(b, err)
		}
	})
	b.Run("Ordered/ReadBig", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		envelopeStorage, clean := factory(b)
		defer clean()

		bigEnvFragments := GenBigEnvFragments()
		envID := bigEnvFragments[0].GetIDAsUUID()
		eventID := bigEnvFragments[0].GetEventIDsAsUUIDSlice()[0]
		for _, fragment := range bigEnvFragments[:len(bigEnvFragments)-1] {
			status, err := envelopeStorage.StoreEnvelope(ctx, fragment)
			assert.Nil(b, err)
			assert.Equal(b, storage.EnvelopeStorageReceptionReceiving, status.Status)
		}

		status, err := envelopeStorage.StoreEnvelope(ctx, bigEnvFragments[len(bigEnvFragments)-1])
		assert.Nil(b, err)
		assert.Equal(b, storage.EnvelopeStorageReceptionComplete, status.Status)
		assert.Equal(b, "", status.Error())

		b.ResetTimer()
		for i := 0; i != b.N; i++ {
			pos := storage.EnvelopeStartPos()
			itemCount := 0
			for !pos.Complete {
				out, newPos, err := envelopeStorage.ReadEnvelope(ctx, envID, pos, 500)
				if err != nil {
					b.Fatal(err)
				}
				for _, ev := range out.Events {
					itemCount += len(ev.Items)
					if newPos.Events[eventID].ItemCount > 202 {
						b.Fatal("ReadEnvelope returning too many items")
					}
				}
				pos = newPos
			}
			assert.Equal(b, 202, itemCount)
		}
	})
	b.Run("Ordered/WriteBig", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		b.StopTimer()
		bigEnvFragments := GenBigEnvFragments()

		for i := 0; i != b.N; i++ {
			envelopeStorage, clean := factory(b)
			b.StartTimer()

			for _, fragment := range bigEnvFragments[:len(bigEnvFragments)-1] {
				status, err := envelopeStorage.StoreEnvelope(ctx, fragment)
				assert.Nil(b, err)
				assert.Equal(b, storage.EnvelopeStorageReceptionReceiving, status.Status)
			}

			status, err := envelopeStorage.StoreEnvelope(ctx, bigEnvFragments[len(bigEnvFragments)-1])
			assert.Nil(b, err)
			assert.Equal(b, storage.EnvelopeStorageReceptionComplete, status.Status)
			assert.Equal(b, "", status.Error())

			b.StopTimer()
			clean()
		}
	})
	b.Run("StoreOnceReadFiveBigEnvelope", func(b *testing.B) {
		ctx := testutils.GetTestContext(b)
		b.StopTimer()
		bigEnvFragments := GenBigEnvFragments()

		envID := bigEnvFragments[0].GetIDAsUUID()
		eventID := bigEnvFragments[0].GetEventIDsAsUUIDSlice()[0]

		for i := 0; i != b.N; i++ {
			envelopeStorage, clean := factory(b)
			b.StartTimer()

			for _, fragment := range bigEnvFragments[:len(bigEnvFragments)-1] {
				status, err := envelopeStorage.StoreEnvelope(ctx, fragment)
				assert.Nil(b, err)
				assert.Equal(b, storage.EnvelopeStorageReceptionReceiving, status.Status)
			}

			status, err := envelopeStorage.StoreEnvelope(ctx, bigEnvFragments[len(bigEnvFragments)-1])
			assert.Nil(b, err)
			assert.Equal(b, storage.EnvelopeStorageReceptionComplete, status.Status)
			assert.Equal(b, "", status.Error())

			for i := 0; i != b.N; i++ {
				pos := storage.EnvelopeStartPos()
				itemCount := 0
				for !pos.Complete {
					out, newPos, err := envelopeStorage.ReadEnvelope(ctx, envID, pos, 500)
					if err != nil {
						b.Fatal(err)
					}
					for _, ev := range out.Events {
						itemCount += len(ev.Items)
						if newPos.Events[eventID].ItemCount > 202 {
							b.Fatal("ReadEnvelope returning too many items")
						}
					}
					pos = newPos
				}
				assert.Equal(b, 202, itemCount)
			}

			b.StopTimer()
			clean()
		}
	})
}
