// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package utils

import (
	"fmt"
	"io"
	"strings"

	jsoniter "github.com/json-iterator/go"
)

type json2YamlContext struct {
	depth      int
	inproperty bool
}

func JSON2YAML(w io.Writer, r io.Reader) error {
	i := jsoniter.ConfigDefault.BorrowIterator(make([]byte, 1024))
	i.Reset(r)
	i.Attachment = &json2YamlContext{}

	return JSONIter2YAML(i, w)
}

func quoteString(s string) string {
	return `"` + strings.ReplaceAll(s, `"`, `\"`) + `"`
}

func JSONIter2YAML(i *jsoniter.Iterator, w io.Writer) error {
	var ctx = i.Attachment.(*json2YamlContext)
	var inproperty = ctx.inproperty
	ctx.inproperty = false
	var writeErr error
	write := func(s string) {
		if writeErr != nil {
			return
		}
		_, writeErr = w.Write([]byte(s))
	}

	switch i.WhatIsNext() {
	case jsoniter.StringValue:
		value := i.ReadString()
		if strings.Contains(value, "\n") {
			splitted := strings.Split(value, "\n")
			for j := range splitted {
				splitted[j] = strings.Repeat("  ", ctx.depth+1) + splitted[j]
			}
			value = "|+\n" + strings.Join(splitted, "\n")
		} else {
			value = quoteString(value)
		}
		if inproperty {
			write(" ")
		}
		write(value)
		write("\n")

	case jsoniter.NumberValue:
		if inproperty {
			write(" ")
		}
		write(i.ReadNumber().String())
		write("\n")

	case jsoniter.NilValue:
		if inproperty {
			write(" ")
		}
		i.ReadNil()
		write("null\n")

	case jsoniter.BoolValue:
		if inproperty {
			write(" ")
		}
		var value = "false"
		if i.ReadBool() {
			value = "true"
		}
		write(value)
		write("\n")

	case jsoniter.ArrayValue:
		var (
			emptyArray = true
			err        error
		)
		if inproperty {
			ctx.depth--
		}
		i.ReadArrayCB(func(i *jsoniter.Iterator) bool {
			if emptyArray {
				emptyArray = false
				if inproperty {
					write("\n")
				}
			}
			write(strings.Repeat("  ", ctx.depth))
			write("- ")
			if writeErr != nil {
				return false
			}
			ctx.depth++
			err = JSONIter2YAML(i, w)
			ctx.depth--
			return err == nil
		})
		if inproperty {
			ctx.depth++
		}
		if err != nil {
			return err
		}
		if emptyArray {
			write("[]\n")
		}
	case jsoniter.ObjectValue:
		var (
			emptyObject = true
			err         error
		)
		i.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
			if emptyObject {
				emptyObject = false
				if inproperty {
					write("\n")
					write(strings.Repeat("  ", ctx.depth))
				}
			} else {
				write(strings.Repeat("  ", ctx.depth))
			}
			write(s)
			write(":")
			if writeErr != nil {
				return false
			}
			ctx.depth++
			ctx.inproperty = true
			err = JSONIter2YAML(i, w)
			ctx.inproperty = false
			ctx.depth--
			return err == nil
		})
	}
	if writeErr != nil && i.Error != nil {
		return fmt.Errorf("JsonIter2Yaml: %w / %w", writeErr, i.Error)
	}
	if writeErr != nil {
		return writeErr
	}
	return i.Error
}
