package demo

import (
	"context"
	"fmt"
	"time"

	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/envelope"
)

// NewHelloWorldEmitter returns a HelloWorldEmitter ActorService
// It is registered as "demo.helloworld" in the ActorService registry
func NewHelloWorldEmitter(actor *xbus.Actor, settings xbus.Settings) (xbus.ActorService, error) {
	return &HelloWorldEmitter{
		actor:    actor,
		message:  settings.MustString("message"),
		interval: settings.MustIntD("interval", 1000),
	}, nil
}

// HelloWorldEmitter sends 'demo.simplemessage' events periodically
type HelloWorldEmitter struct {
	actor    *xbus.Actor
	message  string
	interval int
	ticker   *time.Ticker
	quit     chan bool
}

// Startup initialize the ticker
func (e *HelloWorldEmitter) Startup() error {
	e.ticker = time.NewTicker(time.Duration(e.interval) * time.Millisecond)
	e.quit = make(chan bool)
	go func() {
		count := 0
		for {
			select {
			case <-e.ticker.C:
				count++
				message := fmt.Sprintf("%s - %d", e.message, count)
				e.actor.Log.Info().Str("msg", message).Msg("Sending message")
				if _, err := e.actor.Emit(
					context.Background(),
					envelope.MustNewEnvelope(
						envelope.MustNewTextMessage("demo.simplemessage", message)),
				); err != nil {
					e.actor.Log.Err(err).Msg("send failed")
				}
			case <-e.quit:
				return
			}
		}
	}()
	return nil
}

// Shutdown stops the ticker
func (e *HelloWorldEmitter) Shutdown() error {
	e.ticker.Stop()
	e.ticker = nil
	e.quit <- true
	return nil
}

func init() {
	xbus.RegisterActorService("demo.helloworld", NewHelloWorldEmitter)
}
