package envelope

import (
	"context"
	"errors"
	"io"
	"sync"
	"time"

	"xbus.io/go-xbus/v4/api"
)

// ErrTimeout ...
var ErrTimeout = errors.New("Timeout")

// Complete accumulate the complete envelope as a api.Envelope
func Complete(e Envelope, timeout time.Duration) (*api.Envelope, error) {
	ctx, cancel := context.WithTimeout(context.Background(), timeout)
	defer cancel()
	return CompleteWithContext(ctx, e)
}

// CompleteWithContext accumulate the complete envelope as a api.Envelope
func CompleteWithContext(ctx context.Context, e Envelope) (*api.Envelope, error) {
	done := make(chan struct{})
	defer close(done)
	go func() {
		select {
		case <-ctx.Done():
			_ = e.CloseWithError(ctx.Err())
		case <-done:
		}
	}()
	var (
		result = api.Envelope{
			ID:   e.ID().ToBytes(),
			Last: true,
		}
		finalErr error
		mu       sync.Mutex
		wg       sync.WaitGroup
	)
	for msg, err := e.NextMessage(true); err != io.EOF; msg, err = e.NextMessage(true) {
		mu.Lock()
		result.EventIDs = append(result.EventIDs, msg.ID().ToBytes())
		var event = api.Event{
			ID:    msg.ID().ToBytes(),
			Type:  msg.MsgType(),
			Index: 1,
		}
		result.Events = append(result.Events, &event)
		mu.Unlock()
		wg.Add(1)
		go func(msg Message) {
			defer wg.Done()
			chunks, err := ReadAllChunks(msg.ChunkReader())
			if err != nil {
				mu.Lock()
				if finalErr == nil {
					finalErr = err
					_ = e.CloseWithError(err)
				}
				mu.Unlock()
				return
			}

			mu.Lock()
			event.ItemCount = uint64(len(chunks))
			event.Checksum = msg.Checksum().AsUint32()
			event.Items = chunks
			mu.Unlock()

		}(msg)
	}
	wg.Wait()
	if finalErr != nil {
		return nil, finalErr
	}
	return &result, nil
}
