package envelope

import (
	"bytes"
	"errors"
	"io"

	"xbus.io/go-xbus/v4/api"
)

var (
	// MaxChunkSize is the maximum allowed size for a chunk. If bigger, any
	// api taking a chunk will refuse it
	MaxChunkSize = 1024 * 768

	// ChunkSplitThreadhold is the chunk size above which a paylod will be
	// splitted in chunks
	ChunkSplitThreadhold = 1024 * 250

	// DefaultChunkSize is the chunk size when a content is automatically
	// splitted
	DefaultChunkSize = 1024 * 50

	// ErrCheckum is returned by addChunk if a checksum error is detected
	ErrCheckum = errors.New("Checksum error")

	// ErrClosedEnvelope is returned by addMessage if the envelope is
	// already closed
	ErrClosedEnvelope = errors.New("Envelope is closed")

	// ErrClosedMessage is returned by Message readers/writers if the message is
	// already closed
	ErrClosedMessage = errors.New("Message is closed")
)

// ChunkReader allow reading a messages by chunks as they were written by a
// ChunkWriter
type ChunkReader interface {
	ReadString() (string, error)
	ReadBytes() ([]byte, error)
	DataAvailable() bool
}

// Message is typed and holds some data
type Message interface {
	ID() api.UUID
	MsgType() string
	Checksum() api.Checksum
	// Clone the message. Cannot be called if any content access method was
	// already called
	Clone() Message

	Reader() io.Reader
	ChunkReader() ChunkReader
}

// MessageWriter is a message with added functions to write it then close it
type MessageWriter interface {
	Message
	io.WriteCloser
	WriteChunk([]byte) error
	WriteStringChunk(string) error
}

// MessageCloser is a Message that should be closed.
type MessageCloser interface {
	CloseWithError(err error) error
	Message
}

// ReadAllAsString returns a message full content as a string
func ReadAllAsString(m Message) (string, error) {
	b, err := ReadAll(m)
	return string(b), err
}

// ReadAll returns a message full content as a []byte
func ReadAll(m Message) ([]byte, error) {
	return io.ReadAll(m.Reader())
}

// ReadAllChunks returns all the chunks of a ChunkReader
func ReadAllChunks(reader ChunkReader) ([][]byte, error) {
	var (
		chunks [][]byte
	)

	for {
		chunk, err := reader.ReadBytes()
		if err == io.EOF {
			return chunks, nil
		}
		if err != nil {
			return nil, err
		}
		chunks = append(chunks, chunk)
	}
}

// NewTextMessage creates a message with a simple string content
// The content, if too big, may be splitted in smaller chunks
func NewTextMessage(msgType string, content string) (Message, error) {
	return NewBytesMessage(msgType, []byte(content))
}

// MustNewTextMessage creates a message and panics if a message UUID cannot
// be generated, which should be virtually never
func MustNewTextMessage(msgType string, content string) Message {
	msg, err := NewTextMessage(msgType, content)
	if err != nil {
		panic(err)
	}
	return msg
}

// NewTextMessageWithID creates a message with a simple string content
// The content, if too big, may be splitted in smaller chunks
func NewTextMessageWithID(id api.UUID, msgType string, content string) (Message, error) {
	return NewBytesMessageWithID(id, msgType, []byte(content))
}

// NewBytesMessage creates a message with a simple []byte content
// The content, if too big, may be splitted in smaller chunks
func NewBytesMessage(msgType string, content []byte) (Message, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, err
	}
	return NewBytesMessageWithID(id, msgType, content)
}

// MustNewBytesMessage creates a message and panics if a message UUID cannot
// be generated, which should be virtually never
func MustNewBytesMessage(msgType string, content []byte) Message {
	msg, err := NewBytesMessage(msgType, content)
	if err != nil {
		panic(err)
	}
	return msg
}

// NewBytesMessageWithID creates a message with a simple []byte content
// The content, if too big, may be splitted in smaller chunks
func NewBytesMessageWithID(id api.UUID, msgType string, content []byte) (Message, error) {
	if len(content) > ChunkSplitThreadhold {
		msg := NewStreamMessage(id, msgType)
		msgW := newStreamMessageWriter(msg)
		if _, err := msgW.Write(content); err != nil {
			return nil, err
		}
		if err := msgW.Close(); err != nil {
			return nil, err
		}
		return msg, nil
	}

	checksum := api.Checksum(0).
		Update(id[:16]).
		Update([]byte(msgType)).
		Update(content)

	return &SimpleMessage{id, checksum, msgType, content}, nil
}

// SimpleMessage is a non-streamed implementation of Message
type SimpleMessage struct {
	id       api.UUID
	checksum api.Checksum
	msgType  string
	content  []byte
}

// ID returns the message ID
func (m *SimpleMessage) ID() api.UUID {
	return m.id
}

// Checksum returns the message checksum
func (m *SimpleMessage) Checksum() api.Checksum {
	return m.checksum
}

// MsgType returns the message type
func (m *SimpleMessage) MsgType() string {
	return m.msgType
}

// Clone the message
func (m *SimpleMessage) Clone() Message {
	return &SimpleMessage{id: m.id, msgType: m.msgType, content: m.content}
}

// Reader returns a io.Reader on the content
func (m *SimpleMessage) Reader() io.Reader {
	return bytes.NewBuffer(m.content)
}

type plainChunkReader struct {
	content []byte
	read    bool
}

// ReadString returns the whole content as string, only once
func (pcr *plainChunkReader) ReadString() (string, error) {
	if pcr.read {
		return "", io.EOF
	}
	pcr.read = true
	return string(pcr.content), nil
}

// ReadBytes load the content as bytes, only once. If the buffer is too small
// a io.ErrShortBuffer is returned and no data is copied to the buffer
func (pcr *plainChunkReader) ReadBytes() ([]byte, error) {
	if pcr.read {
		return nil, io.EOF
	}
	pcr.read = true
	return pcr.content, nil
}

func (pcr *plainChunkReader) DataAvailable() bool {
	return !pcr.read
}

// ChunkReader returns a ChunkReader on the content
func (m *SimpleMessage) ChunkReader() ChunkReader {
	return &plainChunkReader{content: m.content}
}

// ForEachChunk calls a function for each chunk of a message
func ForEachChunk(msg Message, fn func([]byte) error) error {
	reader := msg.ChunkReader()
	for chunk, err := reader.ReadBytes(); err != io.EOF; chunk, err = reader.ReadBytes() {
		if err != nil {
			return err
		}
		if err := fn(chunk); err != nil {
			return err
		}
	}
	return nil
}

// ForEachChunkString calls a function for each chunk of a message
func ForEachChunkString(msg Message, fn func(string) error) error {
	reader := msg.ChunkReader()
	for chunk, err := reader.ReadString(); err != io.EOF; chunk, err = reader.ReadString() {
		if err != nil {
			return err
		}
		if err := fn(chunk); err != nil {
			return err
		}
	}
	return nil
}
