package envelope

import (
	"context"
	"fmt"
	"strings"
	"sync"

	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v4/api"
)

// ToFragments convert a envelope.Envelope into a stream of *api.Envelope (aka fragments)
// if the context is canceled, the envelope CloseWithError function is called with the
// error returned by ctx.Err()
func ToFragments(
	ctx context.Context,
	envelope Envelope,
	maxSize int,
	pushFragment func(*api.Envelope),
) error {
	var (
		wg       sync.WaitGroup
		done     = make(chan struct{})
		msgCount int
		errC     = make(chan error)
		finalErr = make(chan error)
		log      = zerolog.Ctx(ctx)
	)

	defer close(done)

	go func() {
		select {
		case <-ctx.Done():
			if err := envelope.CloseWithError(ctx.Err()); err != nil {
				log.Err(err).Msg("envelope.CloseWithError failed")
			}
		case <-done:
		}
	}()

	go func() {
		var (
			errs     []interface{}
			args     []string
			hasError bool
		)
		for err := range errC {
			if err != ctx.Err() {
				if !hasError {
					hasError = true

					if cerr := envelope.CloseWithError(ctx.Err()); cerr != nil {
						errs = append(errs, cerr)
						args = append(args, "%w")
					}

				}

				errs = append(errs, err)
				args = append(args, "%w")
			}
		}
		if len(errs) != 0 {
			finalErr <- fmt.Errorf("ToFragments error(s): "+strings.Join(args, ", "), errs...)
		} else {
			finalErr <- nil
		}
	}()

	var (
		builder = api.NewSyncFragmentsBuilder(envelope.ID(), pushFragment, api.FragmentsBuilderMaxSize(maxSize))
	)

	// Start reading all messages
	if err := ForEachMessage(envelope, func(msg Message) error {
		if err := builder.AddMessage(msg.ID(), msg.MsgType()); err != nil {
			return err
		}
		msgCount++
		wg.Add(1)
		go func() {
			defer wg.Done()
			msgID := msg.ID()
			var count int
			if err := ForEachChunk(msg, func(chunk []byte) error {
				count++
				return builder.AddChunk(msgID, chunk)
			}); err != nil {
				errC <- err
			}

			if err := builder.CloseMessage(msgID, msg.Checksum()); err != nil {
				errC <- err
			}
		}()
		return nil
	}); err != nil {
		return err
	}
	builder.LastMessage()

	wg.Wait()
	close(errC)

	// Signal that we finished reading the envelope
	if err := <-finalErr; err != nil {
		return err
	}

	return ctx.Err()
}
