package xbus

import (
	"errors"
	"fmt"
	"runtime/debug"

	"github.com/rs/zerolog/log"
)

var (
	ErrNoSuchActorServiceType = errors.New("no such actor service type")
)

// ActorService is a service that implements a actor behavior
type ActorService interface {
}

// StartupShutdown is for services that have a startup and a shutdown
type StartupShutdown interface {
	Startup() error
	Shutdown() error
}

// ActorProcessService is a service that has to handle process resquests
type ActorProcessService interface {
	Process(*ActorProcessingContext) error
}

// NewActorServiceFunc is a type of function that creates ActorService instances
type NewActorServiceFunc func(*Actor, Settings) (ActorService, error)

var actorServiceRegistry = make(map[string]NewActorServiceFunc)

// RegisterActorService registers a new ActorService factory function
func RegisterActorService(typename string, init NewActorServiceFunc) {
	_, ok := actorServiceRegistry[typename]
	if ok {
		panic("ActorService '" + typename + "' already registered")
	}
	actorServiceRegistry[typename] = init
}

// UnregisterActorService unregisters a ActorService
func UnregisterActorService(typename string) {
	delete(actorServiceRegistry, typename)
}

// MakeActorService returns a new ActorService, or nil if not registered
func MakeActorService(typename string, actor *Actor) (svc ActorService, returnErr error) {
	f, ok := actorServiceRegistry[typename]
	if !ok {
		return nil, fmt.Errorf("%w: %s", ErrNoSuchActorServiceType, typename)
	}
	defer func() {
		if r := recover(); r != nil {
			if err, ok := r.(error); ok {
				if errors.Is(err, ErrSettings) {
					returnErr = fmt.Errorf(
						"could not start service '%s' for actor '%s': %w", typename, actor.Name, err,
					)
				} else {
					log.Err(err).
						Str("stack", string(debug.Stack())).
						Msg("unexpected panic")
				}
			} else {
				log.Error().Interface("recovered", r).
					Str("stack", string(debug.Stack())).
					Msg("unexpected panic")
			}
		}
	}()
	svc, err := f(actor, actor.Options.Service.Settings)
	if err != nil {
		return nil, fmt.Errorf(
			"could not initialize service '%s' for actor '%s': %w",
			typename, actor.Name, err,
		)
	}
	if svc == nil {
		panic("factory returned nil")
	}
	return svc, nil
}

// RunningService holds a running service and enough context to shut it down
type RunningService struct {
	actorName       string
	svc             ActorService
	startupShutdown StartupShutdown
	sub             *Subscription
	log             Logger
}

// StartupService starts a service
func StartupService(actor *Actor, svc ActorService) (*RunningService, error) {
	log := actor.Log
	rs := RunningService{
		actorName: actor.Name,
		svc:       svc,
		log:       log,
	}
	if ss, ok := svc.(StartupShutdown); ok {
		log.Debug().Msg("Actor service is a StartupShutdown, calling Startup()")
		if err := ss.Startup(); err != nil {
			return nil, err
		}
		rs.startupShutdown = ss
	}
	if ps, ok := svc.(ActorProcessService); ok {
		log.Debug().Msg("Actor service is a ActorProcessService, registering.")
		sub, err := actor.Subscribe(ps.Process)
		if err != nil {
			if err := rs.Shutdown(); err != nil {
				log.Err(err).Msg("Error shutting down actor")
			}
			return nil, err
		}
		rs.sub = sub
	}
	if rs.startupShutdown == nil && rs.sub == nil {
		return nil, fmt.Errorf("Actor %s service (%T) implements no known interface. It cannot be started", actor.Name, svc)
	}
	return &rs, nil
}

// Shutdown stops a running ActorService
func (rs *RunningService) Shutdown() error {
	if rs.sub != nil {
		if err := rs.sub.Unsubscribe(); err != nil {
			rs.log.Err(err).Msg("Error unsubscribing actor")
		}
		rs.sub = nil
	}
	if rs.startupShutdown != nil {
		// TODO report errors during shutdown
		if err := rs.startupShutdown.Shutdown(); err != nil {
			rs.log.Err(err).Msg("Error shutting down actor %s: %s")
		}
		rs.startupShutdown = nil
	}
	return nil
}
