// Copyright 2016 Orus.io. All rights reserved.

package testutils

import (
	"io"
	"os"
	"strings"
	"testing"

	"github.com/rs/zerolog"
	xbus "xbus.io/go-xbus/v4"
)

// GetTestLogger returns a new BusLogger for the passed testing.T
func GetTestLogger(t testing.TB) *TestLogger {
	logger := TestLogger{
		t:              t,
		expectedErrors: nil,
		printToStdout:  os.Getenv("XBUS_TEST_LOG_TO_STDOUT") != "",
		console:        zerolog.ConsoleWriter{Out: os.Stdout},
		testWriter:     zerolog.NewTestWriter(t),
	}
	logger.testWriter.Frame = 8
	logger.Logger, logger.LoggerOutput = xbus.GetLogger(&xbus.LoggerOptions{
		Logtime:     false,
		Debug:       true,
		Trace:       os.Getenv("TEST_TRACE") == "1",
		FinalOutput: &logger,
	})
	logger.XbusLogger = xbus.NewLogger(logger.Logger)
	return &logger
}

// MatchMessage is a function that match a log message
type MatchMessage func(string) bool

// ExpectedError is the definition of an expected error
type ExpectedError struct {
	MatchMessage MatchMessage
	Name         string
}

// ExpectsEqual returns a ExpectedError matching a message exactly
func ExpectsEqual(value string) ExpectedError {
	return ExpectedError{
		MatchMessage: func(text string) bool { return text == value },
		Name:         value,
	}
}

// ExpectsContains returns a ExpectedError matching a partial message
func ExpectsContains(value string) ExpectedError {
	return ExpectedError{
		Name: value,
		MatchMessage: func(text string) bool {
			return strings.Contains(text, value)
		},
	}
}

// TestLogger is a BusLogger for tests
type TestLogger struct {
	zerolog.Logger
	LoggerOutput   io.Writer
	XbusLogger     xbus.Logger
	t              testing.TB
	testWriter     zerolog.TestWriter
	expectedErrors []ExpectedError
	printToStdout  bool
	console        zerolog.ConsoleWriter
	sublogger      *TestLogger
}

func (m *TestLogger) Zero() *zerolog.Logger {
	return &m.Logger
}

func (m *TestLogger) Write(p []byte) (int, error) {
	if m.sublogger != nil {
		return m.sublogger.Write(p)
	}
	return m.testWriter.Write(p)
}

// WriteLevel implements zerolog.LevelWriter interface
func (m *TestLogger) WriteLevel(level zerolog.Level, p []byte) (int, error) {
	if m.sublogger != nil {
		return m.sublogger.WriteLevel(level, p)
	}
	if m.printToStdout {
		if _, err := m.console.Write(p); err != nil {
			m.t.Error("error writing to console:", err)
		}
	}
	switch level {
	case zerolog.WarnLevel, zerolog.InfoLevel, zerolog.DebugLevel, zerolog.TraceLevel:
		return m.testWriter.Write(p)
	case zerolog.ErrorLevel:
		if m.expectedError(string(p)) {
			_, _ = m.testWriter.Write([]byte("(expected) " + string(p)))
		} else {
			m.t.Error(string(p))
		}
	case zerolog.FatalLevel:
		m.t.Fatal(string(p))
	case zerolog.PanicLevel:
		m.t.Log(string(p))
	default:
		panic("Invalid level: " + level.String())
	}
	return len(p), nil
}

func (m *TestLogger) expectedError(msg string) bool {
	for i, expectedError := range m.expectedErrors {
		if expectedError.MatchMessage(msg) {
			m.expectedErrors = append(m.expectedErrors[:i], m.expectedErrors[i+1:]...)
			return true
		}
	}
	return false
}

// SetTest changes the current test, and return a function to cancel the change
func (m *TestLogger) SetTest(t testing.TB) func() {
	previousTest := m.t
	m.t = t
	return func() {
		m.t = previousTest
	}
}

// ExpectsError anticipate an error and make it non-fatal to the test
func (m *TestLogger) ExpectsError(expectedError interface{}) {
	if eE, ok := expectedError.(ExpectedError); ok {
		m.expectedErrors = append(m.expectedErrors, eE)
	} else if strValue, ok := expectedError.(string); ok {
		m.expectedErrors = append(
			m.expectedErrors,
			ExpectsEqual(strValue),
		)
	} else {
		panic("Invalid type for expectedError. Expected a MatchMessage function " +
			"or a string")
	}
}

// VerifyExpectations enforce error expectations
func (m TestLogger) VerifyExpectations() {
	for _, eE := range m.expectedErrors {
		m.t.Errorf("Expected error not received: %s", eE.Name)
	}
}

// SubTest returns a logger for the given test. Until it is cleaned, all the logs
// from the parent logger will be sent to the subtest logger.
// The returned 'clean' function will run VerifyExpectations() on the sublogger
func (m *TestLogger) SubTest(tb testing.TB) (*TestLogger, func()) {
	m.sublogger = GetTestLogger(tb)
	m.sublogger.testWriter.Frame = m.testWriter.Frame + 1

	return m.sublogger, func() {
		m.sublogger.VerifyExpectations()
		m.sublogger = nil
	}
}

// SubTestCapture is like SubTest but does not return the new logger
func (m *TestLogger) SubTestCapture(tb testing.TB) func() {
	_, clean := m.SubTest(tb)

	return clean
}
